import ase.io.ulm as ulm

from gpaw.io import Writer

from gpaw.lcaotddft.observer import TDDFTObserver


class WaveFunctionReader(object):
    def __init__(self, filename, index=None, wfreader=None):
        if index is None:
            self.reader = ulm.Reader(filename)
            tag = self.reader.get_tag()
            if tag != WaveFunctionWriter.ulmtag:
                raise RuntimeError('Unknown tag %s' % tag)
            self.filename = filename
            self._is_main_reader = True
        else:
            self.index = index
            self.reader = wfreader.reader[index]
            self.version = wfreader.version
            self.split = wfreader.split
            self.filename = wfreader.filename
            self._is_main_reader = False

    def __getattr__(self, attr):
        try:
            return getattr(self.reader, attr)
        except AttributeError:
            pass
        except KeyError:  # backwards compatibility with ase-3.19.0
            pass

        # Split reader handling
        if attr == 'wave_functions' and self.split:
            if not hasattr(self, 'splitreader'):
                self.splitreader = ulm.Reader(self.split_filename)
                tag = self.splitreader.get_tag()
                assert tag == WaveFunctionWriter.ulmtag_split
            return getattr(self.splitreader, attr)

        # Compatibility for older versions
        if attr == 'split':
            return False

        if attr == 'split_filename':
            name, ext = tuple(self.filename.rsplit('.', 1))
            if self.version < 3:
                fname = '%s-%06d-%s.%s' % (name, self.niter, self.action, ext)
            else:
                fname = '%s-%06d.%s' % (name, self.index, ext)
            return fname

        raise AttributeError('Attribute %s not defined in version %s' %
                             (repr(attr), repr(self.version)))

    def __len__(self):
        return len(self.reader)

    def __getitem__(self, index):
        return WaveFunctionReader(None, index, self)

    def close(self):
        if hasattr(self, 'splitreader'):
            self.splitreader.close()
            del self.splitreader
        if self._is_main_reader:
            self.reader.close()

    def __del__(self):
        self.close()


class WaveFunctionWriter(TDDFTObserver):
    version = 3
    ulmtag = 'WFW'
    ulmtag_split = ulmtag + 'split'

    def __init__(self, paw, filename, split=False, interval=1):
        TDDFTObserver.__init__(self, paw, interval)
        self.split = split
        if paw.niter == 0:
            self.writer = Writer(filename, paw.world, mode='w',
                                 tag=self.__class__.ulmtag)
            self.writer.write(version=self.__class__.version)
            self.writer.write(split=self.split)
            self.writer.sync()
            self.index = 1
        else:
            # Check the earlier file
            reader = WaveFunctionReader(filename)
            assert reader.version == self.__class__.version
            self.split = reader.split  # Use the earlier split value
            self.index = len(reader)
            reader.close()

            # Append to earlier file
            self.writer = Writer(filename, paw.world, mode='a',
                                 tag=self.__class__.ulmtag)

        if self.split:
            name, ext = tuple(filename.rsplit('.', 1))
            self.split_filename_fmt = name + '-%06d.' + ext

    def _update(self, paw):
        # Write metadata to main writer
        self.writer.write(niter=paw.niter, time=paw.time, action=paw.action)
        if paw.action == 'kick':
            self.writer.write(kick_strength=paw.kick_strength)

        if self.split:
            # Use separate writer for actual data
            filename = self.split_filename_fmt % self.index
            writer = Writer(filename, paw.world, mode='w',
                            tag=self.__class__.ulmtag_split)
        else:
            # Use the same writer for actual data
            writer = self.writer
        w = writer.child('wave_functions')
        paw.wfs.write_wave_functions(w)
        paw.wfs.write_occupations(w)
        if self.split:
            writer.close()
        # Sync the main writer
        self.writer.sync()
        self.index += 1

    def __del__(self):
        self.writer.close()



"""Test functionality to compute Fourier Transforms with PAW corrections"""

# General modules
import numpy as np
import pytest

# Script modules
from ase.units import Ha

from gpaw import GPAW
import gpaw.mpi as mpi
from gpaw.pw.descriptor import PWDescriptor
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.grid_descriptor import GridDescriptor
from gpaw.lfc import LFC
from gpaw.atom.radialgd import AERadialGridDescriptor

from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.localft import (LocalFTCalculator, MicroSetup,
                                   add_total_density, add_LSDA_Bxc)
from gpaw.response.pair_functions import get_pw_coordinates
from gpaw.test.response.test_site_kernels import get_pw_descriptor


# ---------- Test parametrization ---------- #

# 1s orbital radii
testa_a = np.linspace(0.5, 1.5, 10)  # a.u.


def ae_1s_density(r_g, a=1.0):
    """Construct the radial dependence of the density from a 1s orbital on the
    radial grid r_g."""
    assert np.all(r_g >= 0)
    prefactor = 1 / (np.pi * a**3.)
    n_g = prefactor * np.exp(-2. * r_g / a)

    return n_g


def ae_1s_density_plane_waves(pd, R_v, a=1.0):
    """Calculate the plane-wave components of the density from a 1s
    orbital centered at a given position analytically."""
    # List of all plane waves
    G_Gv = np.array([pd.G_Qv[Q] for Q in pd.Q_qG[0]])
    Gnorm_G = np.linalg.norm(G_Gv, axis=1)

    position_prefactor_G = np.exp(-1.j * np.dot(G_Gv, R_v))
    atomcentered_n_G = 1 / (1 + (Gnorm_G * a / 2.)**2.)**2.

    n_G = position_prefactor_G * atomcentered_n_G

    return n_G


# ---------- Actual tests ---------- #

@pytest.mark.response
@pytest.mark.parametrize("a", testa_a)
def test_localft_grid_calculator(a):
    """Test that the LocalGridFTCalculator is able to correctly Fourier
    transform the all-electron density of an 1s orbital."""
    # ---------- Inputs ---------- #

    # Real-space grid
    lc_to_a_ratio = 10  # lattice constant to orbital radii
    N_grid_points = 50**3

    # Plane-wave cutoff
    relative_ecut = 20  # eV relative to a=1

    # Test tolerance
    rtol = 1e-3

    # ---------- Script ---------- #

    # Set up atomic position at the center of the unit cell
    lattice_constant = lc_to_a_ratio * a  # a.u.
    R_v = np.array([lattice_constant, lattice_constant,
                    lattice_constant]) / 2.  # Place atom at the center

    # Set up grid descriptor
    cell_cv = np.array([[lattice_constant, 0., 0.],
                        [0., lattice_constant, 0.],
                        [0., 0., lattice_constant]])
    N_c = np.array([int(N_grid_points**(1 / 3.))] * 3)
    gd = GridDescriptor(N_c, cell_cv=cell_cv, comm=mpi.serial_comm)

    # Set up plane-wave descriptor
    qd = KPointDescriptor(np.array([[0., 0., 0.]]))
    pd = PWDescriptor(relative_ecut / Ha / a**2., gd, complex, qd)

    # Calculate the plane-wave components analytically
    ntest_G = ae_1s_density_plane_waves(pd, R_v, a=a)

    # Calculate the atomic radius at all grid points
    r_vR = gd.get_grid_point_coordinates()
    r_R = np.linalg.norm(r_vR - R_v[:, np.newaxis, np.newaxis, np.newaxis],
                         axis=0)

    # Calculate the all-electron density on the real-space grid
    n_sR = np.array([ae_1s_density(r_R, a=a)])

    # Initialize the LocalGridFTCalculator with an empty ground state adapter
    gs = EmptyGSAdapter()  # hack to pass isinstance in constructor
    context = ResponseContext()
    localft_calc = LocalFTCalculator.from_rshe_parameters(gs, context,
                                                          rshelmax=None)

    # Compute the plane-wave components numerically
    n_G = localft_calc._calculate(pd, n_sR, add_total_density)

    # Check validity of results
    assert np.allclose(n_G, ntest_G, rtol=rtol)


@pytest.mark.response
@pytest.mark.parametrize("a", testa_a)
def test_localft_paw_engine(a):
    """Test that the LocalPAWFTEngine is able to correctly Fourier
    transform the all-electron density of an 1s orbital."""
    # ---------- Inputs ---------- #

    # Real-space grid
    lc_to_a_ratio = 10  # lattice constant to orbital radii
    N_grid_points = 50**3

    # Radial grid (using standard parameters from Li)
    rgd_a = 0.0023570226039551583
    rgd_b = 0.0004528985507246377
    rcut = 2.0  # a.u.

    # Plane-wave cutoff
    relative_ecut = 20  # eV relative to a=1

    # Settings for the expansion in real spherical harmonics
    rshe_params_p = [{},
                     {'rshelmax': 0},  # test that only l=0 contributes
                     {'rshewmin': 1e-8}]  # test coefficient filter

    # Test tolerance
    rtol = 5e-4

    # ---------- Script ---------- #

    # Set up atomic position at the center of the unit cell
    lattice_constant = lc_to_a_ratio * a  # a.u.
    R_v = np.array([lattice_constant, lattice_constant,
                    lattice_constant]) / 2.
    pos_ac = np.array([[0.5, 0.5, 0.5]])  # Relative atomic positions

    # Set up grid descriptor
    cell_cv = np.array([[lattice_constant, 0., 0.],
                        [0., lattice_constant, 0.],
                        [0., 0., lattice_constant]])
    N_c = np.array([int(N_grid_points**(1 / 3.))] * 3)
    gd = GridDescriptor(N_c, cell_cv=cell_cv, comm=mpi.serial_comm)

    # Set up radial grid descriptor extending all the way to the edge of the
    # unit cell
    redge = np.sqrt(3) * lattice_constant / 2.  # center-corner distance
    Ng = int(np.floor(redge / (rgd_a + rgd_b * redge)) + 1)
    rgd = AERadialGridDescriptor(rgd_a, rgd_b, N=Ng)

    # Set up plane-wave descriptor
    qd = KPointDescriptor(np.array([[0., 0., 0.]]))
    pd = PWDescriptor(relative_ecut / Ha / a**2., gd, complex, qd)

    # Calculate the plane-wave components analytically
    ntest_G = ae_1s_density_plane_waves(pd, R_v, a=a)

    # Calculate the pseudo and ae densities on the radial grid
    n_g = ae_1s_density(rgd.r_g, a=a)
    gcut = rgd.floor(rcut)
    nt_g, _ = rgd.pseudize(n_g, gcut)

    # Set up pseudo and ae densities on the Lebedev quadrature
    from gpaw.sphere.lebedev import Y_nL
    Y_nL = Y_nL[:, :9]  # include only s, p and d
    nL = Y_nL.shape[1]
    n_sLg = np.zeros((1, nL, Ng), dtype=float)
    nt_sLg = np.zeros((1, nL, Ng), dtype=float)
    # 1s <=> (l,m) = (0,0) <=> L = 0
    n_sLg[0, 0, :] += np.sqrt(4. * np.pi) * n_g  # Y_0 = 1 / sqrt(4pi)
    nt_sLg[0, 0, :] += np.sqrt(4. * np.pi) * nt_g

    # Calculate the pseudo density on the real-space grid
    # ------------------------------------------------- #
    # Generate splines on for the pseudo density on the radial grid
    spline = rgd.spline(nt_g, l=0, rcut=redge)
    # Use the LocalizedFunctionsCollection to generate pseudo density
    # on the cubic real space grid
    nt_R = gd.zeros()
    lfc = LFC(gd, [[spline]])
    lfc.set_positions(pos_ac)
    lfc.add(nt_R, c_axi=np.sqrt(4. * np.pi))  # Y_0 = 1 / sqrt(4pi)
    nt_sR = np.array([nt_R])

    # Create MicroSetup(s)
    micro_setup = MicroSetup(rgd, Y_nL, n_sLg, nt_sLg)
    micro_setups = [micro_setup]

    gs = EmptyGSAdapter()  # hack to pass isinstance in constructor
    context = ResponseContext()
    for rshe_params in rshe_params_p:
        # Initialize the LocalPAWFTCalculator with an empty gs adapter
        localft_calc = LocalFTCalculator.from_rshe_parameters(gs, context,
                                                              **rshe_params)

        # Compute the plane-wave components numerically
        n_G = localft_calc.engine.calculate(pd, nt_sR, [R_v], micro_setups,
                                            add_total_density)

        # Check validity of numerical results
        assert np.allclose(n_G, ntest_G, rtol=rtol)


@pytest.mark.response
def test_Fe_bxc(gpw_files):
    """Test the symmetry relation

    (B^xc_G)^* = B^xc_-G

    for a real life system with d-electrons (bcc-Fe)."""
    # ---------- Inputs ---------- #

    # Bxc calculation
    ecut = 100

    # ---------- Script ---------- #

    # Bxc calculation

    # Set up calculator and plane-wave descriptor
    calc = GPAW(gpw_files['fe_pw_wfs'], parallel=dict(domain=1))
    atoms = calc.atoms
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext()
    localft_calc = LocalFTCalculator.from_rshe_parameters(gs, context)
    pd0 = get_pw_descriptor(atoms, calc, [0., 0., 0.],
                            ecut=ecut,
                            gammacentered=True)

    Bxc_G = localft_calc(pd0, add_LSDA_Bxc)

    # Part 3: Check symmetry relation
    G1_G, G2_G = get_inversion_pairs(pd0)

    assert np.allclose(np.conj(Bxc_G[G1_G]), Bxc_G[G2_G])


# ---------- Test functionality ---------- #


class EmptyGSAdapter(ResponseGroundStateAdapter):
    # Make an empty subclass to pass isinstance in constructor
    # In a future where the response code has been liberated from GPAW
    # calculator objects, the
    # >>> assert isinstance(gs, ResponseGroundStateAdapter)
    # statements can be deleted, making this class redundant.

    def __init__(self):
        pass


def get_inversion_pairs(pd0):
    """Get all pairs of G-indices which correspond to inverted reciprocal
    lattice vectors G and -G."""
    G_Gc = get_pw_coordinates(pd0)

    G1_G = []
    G2_G = []
    paired_indices = []
    for G1, G1_c in enumerate(G_Gc):
        if G1 in paired_indices:
            continue  # Already paired

        for G2, G2_c in enumerate(G_Gc):
            if np.all(G2_c == -G1_c):
                G1_G.append(G1)
                G2_G.append(G2)
                paired_indices += [G1, G2]
                break

    assert len(np.unique(paired_indices)) == len(G_Gc)

    return G1_G, G2_G



import numpy as np
import pickle

from ase.units import Hartree

from gpaw import GPAW
from gpaw.kpt_descriptor import to1bz
from gpaw.spinorbit import soc_eigenstates
from gpaw.pw.descriptor import PWDescriptor
import gpaw.mpi as mpi


class Unfold:
    """This Class is used to Unfold the Bands of a supercell (SC) calculations
    into a the primitive cell (PC). As a convention (when possible) capital
    letters variables are related to the SC while lowercase ones to the
    PC """

    def __init__(self,
                 name=None,
                 calc=None,
                 M=None,
                 spinorbit=None):

        self.name = name
        self.calc = GPAW(calc, txt=None, communicator=mpi.serial_comm)
        self.M = np.array(M, dtype=float)
        self.spinorbit = spinorbit

        self.gd = self.calc.wfs.gd.new_descriptor()

        self.kd = self.calc.wfs.kd
        if self.calc.wfs.mode == 'pw':
            self.pd = self.calc.wfs.pd
        else:
            self.pd = PWDescriptor(ecut=None, gd=self.gd, kd=self.kd,
                                   dtype=complex)

        self.acell_cv = self.gd.cell_cv
        self.bcell_cv = 2 * np.pi * self.gd.icell_cv

        self.nb = self.calc.get_number_of_bands()

        self.v_Kmsn = None
        if spinorbit:
            if mpi.world.rank == 0:
                print('Calculating spinorbit Corrections')
            self.nb = 2 * self.calc.get_number_of_bands()
            soc = soc_eigenstates(self.calc)
            self.e_mK = soc.eigenvalues().T
            self.v_Kmsn = soc.eigenvectors()
            if mpi.world.rank == 0:
                print('Done with the spinorbit Corrections')

    def get_K_index(self, K):
        """Find the index of a given K."""

        K = np.array([K])
        bzKG = to1bz(K, self.acell_cv)[0]
        iK = self.kd.where_is_q(bzKG, self.kd.bzk_kc)
        return iK

    def get_g(self, iK):
        """Not all the G vectors are relevant for the bands unfolding,
        but only the ones that match the PC reciprocal vectors.
        This function finds the relevant ones."""

        G_Gv_temp = self.pd.get_reciprocal_vectors(q=iK, add_q=False)
        G_Gc_temp = np.dot(G_Gv_temp, np.linalg.inv(self.bcell_cv))

        iG_list = []
        g_list = []
        for iG, G in enumerate(G_Gc_temp):
            a = np.dot(G, np.linalg.inv(self.M).T)
            check = np.abs(a) % 1 < 1e-5
            check2 = np.abs((np.abs(a[np.where(~check)]) % 1) - 1) < 1e-5
            if all(check) or all(check2):
                iG_list.append(iG)
                g_list.append(G)

        return np.array(iG_list), np.array(g_list)

    def get_G_index(self, iK, G, G_list):
        """Find the index of a given G."""

        G_list -= G
        sumG = np.sum(abs(G_list), axis=1)
        iG = np.where(sumG < 1e-5)[0]
        return iG

    def get_eigenvalues(self, iK):
        """Get the list of eigenvalues for a given iK."""

        if not self.spinorbit:
            e_m = self.calc.get_eigenvalues(kpt=iK, spin=0) / Hartree
        else:
            e_m = self.e_mK[:, iK] / Hartree
        return np.array(e_m)

    def get_pw_wavefunctions_k(self, iK):
        """Get the list of Fourier coefficients of the WaveFunction for a
        given iK. For spinors the number of bands is doubled and a spin
        dimension is added."""

        psi_mgrid = get_rs_wavefunctions_k(self.calc, iK, self.spinorbit,
                                           self.v_Kmsn)
        if not self.spinorbit:
            psi_list_mG = []
            for i in range(len(psi_mgrid)):
                psi_list_mG.append(self.pd.fft(psi_mgrid[i], iK))

            psi_mG = np.array(psi_list_mG)
            return psi_mG
        else:
            u0_list_mG = []
            u1_list_mG = []
            for i in range(psi_mgrid.shape[0]):
                u0_list_mG.append(self.pd.fft(psi_mgrid[i, 0], iK))
                u1_list_mG.append(self.pd.fft(psi_mgrid[i, 1], iK))

            u0_mG = np.array(u0_list_mG)
            u1_mG = np.array(u1_list_mG)

            u_mG = np.zeros((len(u0_mG),
                             2,
                             u0_mG.shape[1]), complex)

            u_mG[:, 0] = u0_mG
            u_mG[:, 1] = u1_mG
            return u_mG

    def get_spectral_weights_k(self, k_t):
        r"""Returns the spectral weights for a given k in the PC:

            P_mK(k_t) = \sum_n |<Km|k_t n>|**2

        which can be shown to be equivalent to:

            P_mK(k_t) = \sum_g |C_Km(g+k_t-K)|**2
        """

        K_c, G_t = find_K_from_k(k_t, self.M)
        iK = self.get_K_index(K_c)
        iG_list, g_list = self.get_g(iK)
        gG_t_list = g_list + G_t

        G_Gv = self.pd.get_reciprocal_vectors(q=iK, add_q=False)
        G_Gc = np.dot(G_Gv, np.linalg.inv(self.bcell_cv))

        igG_t_list = []
        for g in gG_t_list:
            igG_t_list.append(self.get_G_index(iK, g.copy(), G_Gc.copy()))

        C_mG = self.get_pw_wavefunctions_k(iK)
        P_m = []
        if not self.spinorbit:
            for m in range(self.nb):
                P = 0.
                norm = np.sum(np.linalg.norm(C_mG[m, :])**2)
                for iG in igG_t_list:
                    P += np.linalg.norm(C_mG[m, iG])**2
                P_m.append(P / norm)
        else:
            for m in range(self.nb):
                P = 0.
                norm = np.sum(np.linalg.norm(C_mG[m, 0, :])**2 +
                              np.linalg.norm(C_mG[m, 1, :])**2)
                for iG in igG_t_list:
                    P += (np.linalg.norm(C_mG[m, 0, iG])**2 +
                          np.linalg.norm(C_mG[m, 1, iG])**2)
                P_m.append(P / norm)

        return np.array(P_m)

    def get_spectral_weights(self, kpoints, filename=None):
        """Collect the spectral weights for the k points in the kpoints list.

        This function is parallelized over k's."""

        Nk = len(kpoints)
        Nb = self.nb

        world = mpi.world
        if filename is None:
            try:
                e_mK, P_mK = pickle.load(open('weights_' + self.name +
                                              '.pckl', 'rb'))
            except IOError:
                e_Km = []
                P_Km = []
                if world.rank == 0:
                    print('Getting EigenValues and Weights')

                e_Km = np.zeros((Nk, Nb))
                P_Km = np.zeros((Nk, Nb))
                myk = range(0, Nk)[world.rank::world.size]
                for ik in myk:
                    k = kpoints[ik]
                    print('kpoint: %s' % k)
                    K_c, G_c = find_K_from_k(k, self.M)
                    iK = self.get_K_index(K_c)
                    e_Km[ik] = self.get_eigenvalues(iK)
                    P_Km[ik] = self.get_spectral_weights_k(k)

                world.barrier()
                world.sum(e_Km)
                world.sum(P_Km)

                e_mK = np.array(e_Km).T
                P_mK = np.array(P_Km).T
                if world.rank == 0:
                    pickle.dump((e_mK, P_mK),
                                open('weights_' + self.name + '.pckl', 'wb'))
        else:
            e_mK, P_mK = pickle.load(open(filename, 'rb'))

        return e_mK, P_mK

    def spectral_function(self, kpts, x, X, points_name, width=0.002,
                          npts=10000, filename=None):
        r"""Returns the spectral function for all the ks in kpoints:

                                              eta / pi

            A_k(e) = \sum_m  P_mK(k) x  ---------------------

                                        (e - e_mk)**2 + eta**2


        at each k-points defined on npts energy points in the range
        [emin, emax]. The width keyword is FWHM = 2 * eta."""

        Nk = len(kpts)
        A_ke = np.zeros((Nk, npts), float)

        world = mpi.world
        e_mK, P_mK = self.get_spectral_weights(kpts, filename)
        if world.rank == 0:
            print('Calculating the Spectral Function')
        emin = np.min(e_mK) - 5 * width
        emax = np.max(e_mK) + 5 * width
        e = np.linspace(emin, emax, npts)

        for ik in range(Nk):
            for ie in range(len(e_mK[:, ik])):
                e0 = e_mK[ie, ik]
                D = (width / 2 / np.pi) / ((e - e0)**2 + (width / 2)**2)
                A_ke[ik] += P_mK[ie, ik] * D
        if world.rank == 0:
            pickle.dump((e * Hartree, A_ke, x, X, points_name),
                        open('sf_' + self.name + '.pckl', 'wb'))
            print('Spectral Function calculation completed!')
        return


def find_K_from_k(k, M):
    """Gets a k vector in scaled coordinates and returns a K vector and the
    unfolding G in scaled Coordinates."""

    KG = np.dot(M, k)
    G = np.zeros(3, dtype=int)

    for i in range(3):
        if KG[i] > 0.5000001:
            G[i] = int(np.round(KG[i]))
            KG[i] -= np.round(KG[i])
        elif KG[i] < -0.4999999:
            G[i] = int(np.round(KG[i]))
            KG[i] += abs(np.round(KG[i]))

    return KG, G


def get_rs_wavefunctions_k(calc, iK, spinorbit=False, v_Kmsn=None):
    """Get the list of WaveFunction for a given iK. For spinors the number of
    bands is doubled and a spin dimension is added."""

    N_c = calc.wfs.gd.N_c
    k_c = calc.wfs.kd.ibzk_kc[iK]
    Nb = calc.wfs.bd.nbands
    Ns = calc.wfs.nspins
    eikr_R = np.exp(-2j * np.pi * np.dot(np.indices(N_c).T,
                                         k_c / N_c).T)

    if calc.wfs.mode == 'lcao' and not calc.wfs.positions_set:
        calc.initialize_positions()

    if not spinorbit:
        psit_mgrid = np.array([calc.wfs.get_wave_function_array(m, iK, 0) *
                               eikr_R for m in range(Nb)])
        return psit_mgrid
    else:
        v_msn = v_Kmsn[iK]
        v0_mn = v_msn[:, 0]
        v1_mn = v_msn[:, 1]

        u0_ngrid = np.array(
            [calc.wfs.get_wave_function_array(n, iK, 0) * eikr_R
             for n in range(Nb)])
        u1_ngrid = np.array(
            [calc.wfs.get_wave_function_array(n, iK, (Ns - 1)) * eikr_R
             for n in range(Nb)])

        u0_mG = np.swapaxes(np.dot(v0_mn, np.swapaxes(u0_ngrid, 0, 2)), 1, 2)
        u1_mG = np.swapaxes(np.dot(v1_mn, np.swapaxes(u1_ngrid, 0, 2)), 1, 2)
        ut_mgrid = np.zeros((len(u0_mG),
                             2,
                             len(u0_mG[0]),
                             len(u0_mG[0, 0]),
                             len(u0_mG[0, 0, 0])), complex)
        ut_mgrid[:, 0] = u0_mG
        ut_mgrid[:, 1] = u1_mG
        return ut_mgrid


def plot_spectral_function(filename, color='blue', eref=None,
                           emin=None, emax=None):
    """Function to plot spectral function corresponding to the bandstructure
    along the kpoints path."""

    try:
        e, A_ke, x, X, points_name = pickle.load(open(filename + '.pckl',
                                                      'rb'))
    except IOError:
        print('You Need to Calculate the SF first!')
        raise SystemExit()

    import matplotlib.pyplot as plt
    print('Plotting Spectral Function')

    if eref is not None:
        e -= eref
    if emin is None:
        emin = e.min()
    if emax is None:
        emax = e.max()

    A_ke /= np.max(A_ke)
    A_ek = A_ke.T
    A_ekc = np.reshape(A_ek, (A_ek.shape[0], A_ek.shape[1]))

    mycmap = make_colormap(color)

    plt.figure()

    plt.plot([0, x[-1]], 2 * [0.0], '--', c='0.5')
    plt.imshow(A_ekc + 0.23,
               cmap=mycmap,
               aspect='auto',
               origin='lower',
               vmin=0.,
               vmax=1,
               extent=[0, x[-1], e.min(), e.max()])

    for k in X[1:-1]:
        plt.plot([k, k], [emin, emax], lw=0.5, c='0.5')
    plt.xticks(X, points_name, size=20)
    plt.yticks(size=20)
    plt.ylabel('E(eV)', size=20)
    plt.axis([0, x[-1], emin, emax])
    plt.savefig(filename + '_spec.png')
    plt.show()


def plot_band_structure(e_mK, P_mK, x, X, points_name,
                        weights_mK=None, color='red', fit=True, nfit=200):
    """Function to plot the bandstructure using the P_mK weights directly.
    each point is represented with a filled circle, whose size and color
    vary with as a function of P_mK."""

    import matplotlib.pyplot as plt
    print('Plotting Bands Structure')
    emin = e_mK.min()
    emax = e_mK.max()

    new_cmap = make_colormap(color)

    if weights_mK is None:
        weights_mK = P_mK.copy()
    else:
        weights_mK *= P_mK.copy()

    plt.figure()
    plt.plot([0, x[-1]], 2 * [0.0], '--', c='0.5')
    plt.scatter(np.tile(x, len(e_mK)), e_mK.reshape(-1),
                c=P_mK.reshape(-1),
                cmap=new_cmap,
                vmin=0.,
                vmax=1.,
                s=20. * weights_mK.reshape(-1),
                marker='o',
                edgecolor='none')

    for k in X[1:-1]:
        plt.plot([k, k], [emin, emax], lw=0.5, c='0.5')
    plt.xticks(X, points_name, size=20)
    plt.yticks(size=20)
    plt.ylabel('E(eV)', size=20)
    plt.axis([0, x[-1], emin, emax])
    plt.show()


def make_colormap(main_color):
    """Custom colormaps used in plot_spectral function and
    plot_band_structure."""

    from matplotlib.colors import LinearSegmentedColormap
    if main_color == 'blue':
        cdict = {'red': ((0.0, 1.0, 1.0),
                         (0.25, 1.0, 1.0),
                         (0.5, 0.0, 0.0),
                         (1.0, 0.0, 0.0)),

                 'green': ((0.0, 1.0, 1.0),
                           (0.25, 1.0, 1.0),
                           (0.5, 0.0, 0.0),
                           (1.0, 0.0, 0.0)),

                 'blue': ((0.0, 1.0, 1.0),
                          (0.25, 1.0, 1.0),
                          (0.5, 1.0, 1.0),
                          (1.0, 0.75, 0.75))}

    elif main_color == 'red':
        cdict = {'red': ((0.0, 1.0, 1.0),
                         (0.25, 1.0, 1.0),
                         (0.5, 1.0, 1.0),
                         (1.0, 0.75, 0.75)),

                 'green': ((0.0, 1.0, 1.0),
                           (0.25, 1.0, 1.0),
                           (0.5, 0.0, 0.0),
                           (1.0, 0.0, 0.0)),

                 'blue': ((0.0, 1.0, 1.0),
                          (0.25, 1.0, 1.0),
                          (0.5, 0.0, 0.0),
                          (1.0, 0.0, 0.0))}

    elif main_color == 'green':
        cdict = {'red': ((0.0, 1.0, 1.0),
                         (0.25, 1.0, 1.0),
                         (0.5, 0.0, 0.0),
                         (1.0, 0.0, 0.0)),

                 'green': ((0.0, 1.0, 1.0),
                           (0.25, 1.0, 1.0),
                           (0.5, 1.0, 1.0),
                           (1.0, 0.75, 0.75)),

                 'blue': ((0.0, 1.0, 1.0),
                          (0.25, 1.0, 1.0),
                          (0.5, 0.0, 0.0),
                          (1.0, 0.0, 0.0)),

                 'alpha': ((0.0, 0.0, 0.0),
                           (0.25, 0.1, 0.1),
                           (0.5, 1.0, 1.0),
                           (1.0, 1.0, 1.0))}

    cmap = LinearSegmentedColormap('mymap', cdict)
    return cmap


def get_vacuum_level(calc, plot_pot=False):
    """Get the vacuum energy level from a given calculator."""

    calc.restore_state()
    if calc.wfs.mode == 'pw':
        vHt_g = calc.hamiltonian.pd3.ifft(calc.hamiltonian.vHt_q) * Hartree
    else:
        vHt_g = calc.hamiltonian.vHt_g * Hartree
    vHt_z = np.mean(np.mean(vHt_g, axis=0), axis=0)

    if plot_pot:
        import matplotlib.pyplot as plt
        plt.plot(vHt_z)
        plt.show()
    return vHt_z[0]



import pytest
import numpy as np
from ase import Atoms

import gpaw.dscf as dscf
from gpaw import GPAW, MixerSum
from gpaw.mpi import rank


@pytest.mark.skip(reason='TODO')
def test_dscf_forces():
    L = 10.0

    eigensolver = 'cg'
    xc = 'PBE'

    E0 = []
    E1 = []
    d0 = 1.4
    delta = 0.05
    for i in range(3):
        d = d0 + (i - 1) * delta
        atoms = Atoms('N2',
                      [[0.5 * L, 0.5 * L, 0.5 * L - 0.5 * d],
                       [0.5 * L, 0.5 * L, 0.5 * L + 0.5 * d]])
        atoms.set_cell([L, L, L])

        calc_gs = GPAW(h=0.2,
                       nbands=-5,
                       xc=xc,
                       occupations={'name': 'fermi-dirac', 'width': 0.05},
                       eigensolver=eigensolver,
                       spinpol=True,
                       txt=None,
                       mixer=MixerSum(beta=0.1, nmaxold=5, weight=50.0),
                       convergence={'energy': 100,
                                    'density': 100,
                                    'eigenstates': 1.0e-9,
                                    'bands': -1})
        atoms.calc = calc_gs

        E0.append(atoms.get_potential_energy())
        if i == 1:
            F0 = atoms.get_forces()

        calc_es = GPAW(h=0.2,
                       nbands=-5,
                       xc=xc,
                       width=0.05,
                       eigensolver=eigensolver,
                       spinpol=True,
                       txt=None,
                       mixer=MixerSum(beta=0.1, nmaxold=5, weight=50.0),
                       convergence={'energy': 100,
                                    'density': 100,
                                    'eigenstates': 1.0e-9,
                                    'bands': -1})

        n = 5  # LUMO
        molecule = [0, 1]
        wf_u = [kpt.psit_nG[n] for kpt in calc_gs.wfs.kpt_u]
        p_uai = [dict([(molecule[a], P_ni[n])
                       for a, P_ni in kpt.P_ani.items()])
                 for kpt in calc_gs.wfs.kpt_u]

        atoms.calc = calc_es
        lumo = dscf.AEOrbital(calc_es, wf_u, p_uai)
        dscf.dscf_calculation(calc_es, [[1.0, lumo, 1]], atoms)

        E1.append(atoms.get_potential_energy())
        if i == 1:
            F1 = atoms.get_forces()

    f0 = np.sqrt(((F0[1, :] - F0[0, :])**2).sum()) * 0.5
    f0b = (E0[1] - E0[0]) / delta     # backward gradient
    f0f = (E0[2] - E0[1]) / delta     # forward gradient
    if rank == 0:
        print('Ground state')
        print(E0)
        print(f0b, '<', f0, '<', f0f)
    assert f0 > f0b
    assert f0 < f0f

    f1 = np.sqrt(((F1[1, :] - F1[0, :])**2).sum()) * 0.5
    f1b = (E1[1] - E1[0]) / delta     # backward gradient
    f1f = (E1[2] - E1[1]) / delta     # forward gradient
    if rank == 0:
        print('Excited state')
        print(E1)
        print(f1b, '<', f1, '<', f1f)
    assert f1 > f1b
    assert f1 < f1f



import sys
import re
import numpy as np
import datetime


def write_parallel_cube(basename, data, gd, atoms, rank):
    hx = gd.h_cv[0, 0]
    hy = gd.h_cv[1, 1]
    hz = gd.h_cv[2, 2]

    f = open('%s_%06d.cube' % (basename, rank), 'w', 256 * 1024)

    f.write(
        ('GPAW Global (natom  nx ny nz  hx hy hz):  '
         '%6d  %6d %6d %6d  %12.6lf %12.6lf %12.6lf\n')
        % (len(atoms), gd.N_c[0], gd.N_c[1], gd.N_c[2], hx, hy, hz))
    f.write(
        ('GPAW Local (xbeg xend  ybeg yend  zbeg zend):  '
         '%6d %6d  %6d %6d  %6d %6d\n')
        % (gd.beg_c[0], gd.end_c[0], gd.beg_c[1], gd.end_c[1], gd.beg_c[2],
           gd.end_c[2]))

    f.write('%6d %12.6lf %12.6lf %12.6lf\n' %
            (len(atoms), hx * gd.beg_c[0], hy * gd.beg_c[1], hz * gd.beg_c[2]))
    f.write('%6d %12.6lf %12.6lf %12.6lf\n' % (gd.n_c[0], hx, 0.0, 0.0))
    f.write('%6d %12.6lf %12.6lf %12.6lf\n' % (gd.n_c[1], 0.0, hy, 0.0))
    f.write('%6d %12.6lf %12.6lf %12.6lf\n' % (gd.n_c[2], 0.0, 0.0, hz))

    for (i, atom) in enumerate(atoms):
        f.write('%6d %12.6lf %12.6lf %12.6lf %12.6lf\n' %
                (atom.number, 0.0, atom.position[0] / 0.529177,
                 atom.position[1] / 0.529177, atom.position[2] / 0.529177))
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            for k in range(data.shape[2]):
                f.write('%18.8lf\n' % data[i, j, k])
    f.close()


def cubify(out_filename, filenames):
    sys.stderr.write('Reading partial cube files...%s\n' %
                     datetime.datetime.now())
    data = None
    for fname in filenames:
        f = open(fname, 'r', 256 * 1024)
        line0 = f.readline()
        elems = re.compile(r'[\d.e+-]+').findall(line0)
        natom = int(elems[0])
        (nx, ny, nz) = [int(n) for n in elems[1:4]]
        (hx, hy, hz) = [float(h) for h in elems[4:7]]
        line = f.readline()
        (xb, xe, yb, ye, zb,
         ze) = [int(x) for x in re.compile(r'\d+').findall(line)]
        sys.stderr.write('%d %d %d %d %d %d\n' % (xb, xe, yb, ye, zb, ze))

        line = f.readline()
        line = f.readline()
        line = f.readline()
        line = f.readline()

        atom_lines = []
        for n in range(natom):
            atom_lines.append(f.readline())

        if data is None:
            data = np.zeros((nx, ny, nz))

        for i in range(xb, xe):
            for j in range(yb, ye):
                for k in range(zb, ze):
                    data[i, j, k] = float(f.readline())

        f.close()

    sys.stderr.write('Reading done. %s\n' % datetime.datetime.now())
    sys.stderr.write('Writing the cube file... %d %d %d\n' % (nx, ny, nz))

    out = open(out_filename, 'w', 256 * 1024)
    out.write(line0)
    out.write('OUTER LOOP: X, MIDDLE LOOP: Y, INNER LOOP: Z\n')
    out.write('%8d %12.6lf %12.6lf %12.6lf\n' % (natom, 0.0, 0.0, 0.0))
    out.write('%8d %12.6lf %12.6lf %12.6lf\n' % (nx, hx, 0.0, 0.0))
    out.write('%8d %12.6lf %12.6lf %12.6lf\n' % (ny, 0.0, hy, 0.0))
    out.write('%8d %12.6lf %12.6lf %12.6lf\n' % (nz, 0.0, 0.0, hz))
    for n in range(natom):
        out.write(atom_lines[n])
    for i in range(nx):
        sys.stderr.write('.')
        for j in range(ny):
            for k in range(nz):
                out.write('%18.8lf\n' % data[i, j, k])
    out.close()
    sys.stderr.write('\n')
    sys.stderr.write('Writing done. %s\n' % datetime.datetime.now())


def isocubes(filename, isoval, scale):
    """Requires PIL"""
    import Image

    f = open(filename, 'r', 256 * 1024)
    f.readline()
    f.readline()

    natom = int(f.readline().split()[0])

    elems = f.readline().split()
    nx = int(elems[0])
    # x0 = float(elems[1])

    elems = f.readline().split()
    ny = int(elems[0])
    # y0 = float(elems[2])

    elems = f.readline().split()
    nz = int(elems[0])
    # z0 = float(elems[3])

    for i in range(natom):
        f.readline()

    # data = np.zeros((nx,ny,nz))
    dataxy = [[0.0, 0.0] for i in range(nx * ny)]
    dataxz = [[0.0, 0.0] for i in range(nx * nz)]
    datayz = [[0.0, 0.0] for i in range(ny * nz)]
    for i in range(nx):
        for j in range(ny):
            for k in range(nz):
                val = float(f.readline())
                if val > isoval:
                    dataxy[i * ny + j][0] += val * scale
                    dataxy[i * ny + j][1] += val * scale
                    dataxz[i * nz + k][0] += val * scale
                    dataxz[i * nz + k][1] += val * scale
                    datayz[j * nz + k][0] += val * scale
                    datayz[j * nz + k][1] += val * scale
                elif -val > isoval:
                    dataxy[i * ny + j][0] += -val * scale
                    dataxy[i * ny + j][1] -= -val * scale
                    dataxz[i * nz + k][0] += -val * scale
                    dataxz[i * nz + k][1] -= -val * scale
                    datayz[j * nz + k][0] += -val * scale
                    datayz[j * nz + k][1] -= -val * scale

    image_xy = Image.new('RGB', (nx, ny))
    image_xy_data = image_xy.load()

    for i in range(nx):
        for j in range(ny):
            n = dataxy[i * ny + j][0] * 100
            d = dataxy[i * ny + j][1] * 100
            if n > 255:
                n = 255
                d = 255 * d / n
            n = int(n + .5)
            d = int(d)
            if d > 0:
                dataxy[i * ny + j] = (n, n - d, n - d)
            elif d < 0:
                dataxy[i * ny + j] = (n + d, n + d, n)
            else:
                dataxy[i * ny + j] = (0, 0, 0)

            image_xy_data[i, j] = dataxy[i * ny + j]

    image_xy.save(filename + '_iso_xy.png')

    image_xz = Image.new('RGB', (nx, nz))
    image_xz_data = image_xz.load()

    for i in range(nx):
        for j in range(nz):
            n = dataxz[i * nz + j][0] * 100
            d = dataxz[i * nz + j][1] * 100
            if n > 255:
                n = 255
                d = 255 * d / n
            n = int(n + .5)
            d = int(d)
            if d > 0:
                dataxz[i * nz + j] = (n, n - d, n - d)
            elif d < 0:
                dataxz[i * nz + j] = (n + d, n + d, n)
            else:
                dataxz[i * nz + j] = (0, 0, 0)

            image_xz_data[i, j] = dataxz[i * nz + j]

    image_xz.save(filename + '_iso_xz.png')

    image_yz = Image.new('RGB', (ny, nz))
    image_yz_data = image_yz.load()

    for i in range(ny):
        for j in range(nz):
            n = datayz[i * nz + j][0] * 100
            d = datayz[i * nz + j][1] * 100
            if n > 255:
                n = 255
                d = 255 * d / n
            n = int(n + .5)
            d = int(d)
            if d > 0:
                datayz[i * nz + j] = (n, n - d, n - d)
            elif d < 0:
                datayz[i * nz + j] = (n + d, n + d, n)
            else:
                datayz[i * nz + j] = (0, 0, 0)

            image_yz_data[i, j] = datayz[i * nz + j]

    image_yz.save(filename + '_iso_yz.png')



"""Hyperfine parameters.

See:

    First-principles calculations of defects in oxygen-deficient
    silica exposed to hydrogen

    Peter E. Blöchl

    Phys. Rev. B 62, 6158 – Published 1 September 2000

    https://doi.org/10.1103/PhysRevB.62.6158

"""
from math import pi
from typing import List

import ase.units as units
import numpy as np
from scipy.integrate import simps

from gpaw.calculator import GPAW
from gpaw.atom.aeatom import Channel
from gpaw.atom.configurations import configurations
from gpaw.atom.radialgd import RadialGridDescriptor
from gpaw.gaunt import gaunt
from gpaw.grid_descriptor import GridDescriptor
from gpaw.setup import Setup
from gpaw.typing import Array1D, Array2D, Array3D
from gpaw.utilities import unpack2
from gpaw.pw.descriptor import PWDescriptor
from gpaw.xc.functional import XCFunctional

# Fine-structure constant: (~1/137)
alpha = 0.5 * units._mu0 * units._c * units._e**2 / units._hplanck

g_factor_e = 2.00231930436256


def hyperfine_parameters(calc: GPAW,
                         exclude_core=False) -> Array3D:
    r"""Calculate isotropic and anisotropic hyperfine coupling paramters.

    One tensor (:math:`A_{ij}`) per atom is returned in eV units.
    In Hartree atomic units, we have the isotropic part
    :math:`a = \text{Tr}(\mathbf{A}) / 3`:

    .. math::

        a = \frac{2 \alpha^2 g_e m_e}{3 m_p}
            \int \delta_T(\mathbf{r}) \rho_s(\mathbf{r}) d\mathbf{r},

    and the anisotropic part :math:`\mathbf{A} - a`:

    .. math::

        \frac{\alpha^2 g_e m_e}{4 \pi m_p}
        \int \frac{3 r_i r_j - \delta_{ij} r^2}{r^5}
        \rho_s(\mathbf{r}) d\mathbf{r}.

    Remember to multiply each tensor by the g-factors of the nuclei.

    Use ``exclude_core=True`` to exclude contribution from "frozen" core.
    """
    dens = calc.density
    nt_sR = dens.nt_sG
    A_avv = smooth_part(
        nt_sR[0] - nt_sR[1],
        dens.gd,
        calc.atoms.get_scaled_positions())

    D_asp = calc.density.D_asp
    for a, D_sp in D_asp.items():
        density_sii = unpack2(D_sp)
        setup = calc.wfs.setups[a]

        A_vv = paw_correction(density_sii,
                              setup,
                              calc.hamiltonian.xc,
                              exclude_core)

        A_avv[a] += A_vv

    A_avv *= pi * alpha**2 * g_factor_e * units._me / units._mp * units.Ha

    return A_avv


def smooth_part(spin_density_R: Array3D,
                gd: GridDescriptor,
                spos_ac: Array2D,
                ecut: float = None) -> Array3D:
    """Contribution from pseudo spin-density."""
    pd = PWDescriptor(ecut, gd)
    spin_density_G = pd.fft(spin_density_R)
    G_Gv = pd.get_reciprocal_vectors(add_q=False)
    # eiGR_aG = np.exp(-1j * spos_ac.dot(gd.cell_cv).dot(G_Gv.T))
    eiGR_aG = np.exp(-1j * spos_ac @ gd.cell_cv @ G_Gv.T)

    # Isotropic term:
    W1_a = pd.integrate(spin_density_G, eiGR_aG) / gd.dv * (2 / 3)

    spin_density_G[0] = 0.0
    G2_G = pd.G2_qG[0].copy()
    G2_G[0] = 1.0
    spin_density_G /= G2_G

    # Anisotropic term:
    W_vva = np.empty((3, 3, len(spos_ac)))
    for v1 in range(3):
        for v2 in range(3):
            W_a = pd.integrate(G_Gv[:, v1] * G_Gv[:, v2] * spin_density_G,
                               eiGR_aG)
            W_vva[v1, v2] = -W_a / gd.dv

    W_a = np.trace(W_vva) / 3
    for v in range(3):
        W_vva[v, v] -= W_a
        W_vva[v, v] += W1_a

    return W_vva.transpose((2, 0, 1))


# Normalization constants for xy, yz, 3z^2-r^2, xz, x^2-y^2:
Y2_m = (np.array([15 / 4,
                  15 / 4,
                  5 / 16,
                  15 / 4,
                  15 / 16]) / pi)**0.5
# Second derivatives:
Y2_mvv = np.array([[[0, 1, 0],
                    [1, 0, 0],
                    [0, 0, 0]],
                   [[0, 0, 0],
                    [0, 0, 1],
                    [0, 1, 0]],
                   [[-2, 0, 0],
                    [0, -2, 0],
                    [0, 0, 4]],
                   [[0, 0, 1],
                    [0, 0, 0],
                    [1, 0, 0]],
                   [[2, 0, 0],
                    [0, -2, 0],
                    [0, 0, 0]]])


def paw_correction(density_sii: Array3D,
                   setup: Setup,
                   xc: XCFunctional = None,
                   exclude_core: bool = False) -> Array2D:
    """Corrections from 1-center expansions of spin-density."""
    # Spherical part:
    spin_density_ii = density_sii[0] - density_sii[1]
    D0_jj = expand(spin_density_ii, setup.l_j, l=0)[0]

    phit_jg = np.array(setup.data.phit_jg)
    phi_jg = np.array(setup.data.phi_jg)

    rgd = setup.rgd

    # Spin-density from pseudo density:
    nt0 = phit_jg[:, 0].dot(D0_jj).dot(phit_jg[:, 0]) / (4 * pi)**0.5

    # All-electron contribution diverges as r^-beta and must be integrated
    # over a small region of size rT:
    n0_g = np.einsum('ab, ag, bg -> g', D0_jj, phi_jg, phi_jg) / (4 * pi)**0.5
    if not exclude_core and setup.Nc > 0 and xc is not None:
        n0_g += core_contribution(density_sii, setup, xc)

    beta = 2 * (1 - (1 - (setup.Z * alpha)**2)**0.5)
    rT = setup.Z * alpha**2
    n0 = integrate(n0_g, rgd, rT, beta)

    W1 = (n0 - nt0) * 2 / 3  # isotropic result

    # Now the anisotropic part from the l=2 part of the density:
    D2_mjj = expand(spin_density_ii, setup.l_j, 2)
    dn2_mg = np.einsum('mab, ag, bg -> mg', D2_mjj, phi_jg, phi_jg)
    dn2_mg -= np.einsum('mab, ag, bg -> mg', D2_mjj, phit_jg, phit_jg)
    A_m = dn2_mg[:, 1:].dot(rgd.dr_g[1:] / rgd.r_g[1:]) / 5
    A_m *= Y2_m
    # W_vv: Array2D = Y2_mvv.T.dot(A_m)  # type: ignore
    W_vv = Y2_mvv.T @ A_m
    W = np.trace(W_vv) / 3
    for v in range(3):
        W_vv[v, v] -= W
        W_vv[v, v] += W1

    return W_vv


def expand(D_ii: Array2D,
           l_j: List[int],
           l: int) -> Array3D:
    """Get expansion coefficients."""
    G_LLm = gaunt(lmax=2)[:, :, l**2:(l + 1)**2]
    D_mjj = np.empty((2 * l + 1, len(l_j), len(l_j)))
    i1a = 0
    for j1, l1 in enumerate(l_j):
        i1b = i1a + 2 * l1 + 1
        i2a = 0
        for j2, l2 in enumerate(l_j):
            i2b = i2a + 2 * l2 + 1
            D_mjj[:, j1, j2] = np.einsum('ab, abm -> m',
                                         D_ii[i1a:i1b, i2a:i2b],
                                         G_LLm[l1**2:(l1 + 1)**2,
                                               l2**2:(l2 + 1)**2])
            i2a = i2b
        i1a = i1b
    return D_mjj


def delta(r: Array1D, rT: float) -> Array1D:
    """Extended delta function."""
    return 2 / rT / (1 + 2 * r / rT)**2


def integrate(n0_g: Array1D,
              rgd: RadialGridDescriptor,
              rT: float,
              beta: float) -> float:
    """Take care of r^-beta divergence."""
    r_g = rgd.r_g
    a_i = np.polyfit(r_g[1:5], n0_g[1:5] * r_g[1:5]**beta, 3)
    r4 = r_g[4]
    dr = rT / 50
    n = max(int(r4 / dr / 2) * 2 + 1, 3)
    r_j = np.linspace(0, r4, n)

    b_i = np.arange(3, -1, -1) + 1 - beta
    d0 = 2 / rT  # delta(0, rT)
    d1 = -8 / rT**2
    n0 = a_i.dot(d0 * r4**b_i / b_i + d1 * r4**(b_i + 1) / (b_i + 1))

    d_j = delta(r_j, rT) - (d0 + d1 * r_j)

    head_j = d_j * np.polyval(a_i, r_j)
    head_j[1:] *= r_j[1:]**-beta
    n0 += simps(head_j, r_j)

    tail_g = n0_g[4:] * delta(r_g[4:], rT)
    n0 += simps(tail_g, r_g[4:], even='first')

    return n0


def core_contribution(density_sii: Array3D,
                      setup: Setup,
                      xc: XCFunctional) -> Array1D:
    """Calculate spin-density from "frozen" core."""
    # Spherical part:
    D_sjj = [expand(density_ii, setup.l_j, 0)[0]
             for density_ii in density_sii]
    phi_jg = np.array(setup.data.phi_jg)
    rgd = setup.rgd

    # Densities with frozen core:
    n_sg = np.einsum('ag, sab, bg -> sg',
                     phi_jg, D_sjj, phi_jg) / (4 * pi)**0.5
    n_sg += setup.data.nc_g * (0.5 / (4 * pi)**0.5)

    # Potential:
    v_sg = np.zeros_like(n_sg)
    xc.calculate_spherical(rgd, n_sg, v_sg)
    vr_sg = v_sg * rgd.r_g
    vr_sg -= setup.Z
    vr_sg += rgd.poisson(n_sg.sum(axis=0))

    # Find first bound s-state included in PAW potential:
    for n0, l in zip(setup.n_j, setup.l_j):
        if l == 0:
            assert n0 > 0
            break
    else:
        assert False, (setup.n_j, setup.l_j)

    # Initial guesses for core s-states:
    eigs = [e for n, l, f, e in configurations[setup.symbol][1]
            if n < n0 and l == 0]

    # Solve spherical scalar-relativistic Schrödinger equation:
    core_spin_density_g = rgd.zeros()
    sign = 1.0
    for vr_g in vr_sg:
        channel = Channel(l=0, f_n=[1] * len(eigs))
        channel.e_n = eigs
        channel.phi_ng = rgd.empty(len(eigs))
        channel.solve2(vr_g, rgd=rgd, scalar_relativistic=True, Z=setup.Z)
        assert channel.solve2ok
        core_spin_density_g += sign * channel.calculate_density()
        sign = -1.0

    return core_spin_density_g


# From https://en.wikipedia.org/wiki/Gyromagnetic_ratio
# Units: MHz/T
gyromagnetic_ratios = {'H': (1, 42.577478518),
                       'He': (3, -32.434),
                       'Li': (7, 16.546),
                       'C': (13, 10.7084),
                       'N': (14, 3.077),
                       'O': (17, -5.772),
                       'F': (19, 40.052),
                       'Na': (23, 11.262),
                       'Al': (27, 11.103),
                       'Si': (29, -8.465),
                       'P': (31, 17.235),
                       'Fe': (57, 1.382),
                       'Cu': (63, 11.319),
                       'Zn': (67, 2.669),
                       'Xe': (129, -11.777)}


def main(argv: List[str] = None) -> None:
    import argparse
    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.hyperfine',
        description='Calculate hyperfine parameters.')
    add = parser.add_argument
    add('file', metavar='input-file',
        help='GPW-file (with or without wave functions).')
    add('-g', '--g-factors',
        help='G-factors.  Example: "-g H:5.6,O:-0.76".')
    add('-u', '--units', default='ueV', choices=['ueV', 'MHz'],
        help='Units.  Must be "ueV" (micro-eV, default) or "MHz".')
    add('-x', '--exclude-core', action='store_true')
    add('-d', '--diagonalize', action='store_true',
        help='Show eigenvalues of tensor.')

    args = parser.parse_intermixed_args(argv)

    calc = GPAW(args.file)
    atoms = calc.get_atoms()

    symbols = atoms.symbols
    magmoms = atoms.get_magnetic_moments()
    total_magmom = atoms.get_magnetic_moment()

    g_factors = {symbol: ratio * 1e6 * 4 * pi * units._mp / units._e
                 for symbol, (n, ratio) in gyromagnetic_ratios.items()}

    if args.g_factors:
        for symbol, value in (part.split(':')
                              for part in args.g_factors.split(',')):
            g_factors[symbol] = float(value)

    if args.units == 'ueV':
        scale = 1e6
        unit = 'μeV'
    else:
        scale = units._e / units._hplanck * 1e-6
        unit = 'MHz'

    A_avv = hyperfine_parameters(calc, args.exclude_core)

    print('Hyperfine coupling paramters '
          f'in {unit}:\n')

    if args.diagonalize:
        columns = ['1.', '2.', '3.']
    else:
        columns = ['xx', 'yy', 'zz', 'yz', 'xz', 'xy']
    print('  atom  magmom      ', '       '.join(columns))

    used = {}
    for a, A_vv in enumerate(A_avv):
        symbol = symbols[a]
        magmom = magmoms[a]
        g_factor = g_factors.get(symbol, 1.0)
        used[symbol] = g_factor
        A_vv *= g_factor * scale
        if args.diagonalize:
            numbers = np.linalg.eigvalsh(A_vv).tolist()
        else:
            numbers = [A_vv[0, 0], A_vv[1, 1], A_vv[2, 2],
                       A_vv[1, 2], A_vv[0, 2], A_vv[0, 1]]
        print(f'{a:3} {symbol:>2}  {magmom:6.3f}',
              ''.join(f'{x:9.2f}' for x in numbers))

    print('\nCore correction',
          'NOT included!' if args.exclude_core else 'included')
    print(f'Total magnetic moment: {total_magmom:.3f}')

    print('\nG-factors used:')
    for symbol, g in used.items():
        print(f'{symbol:2} {g:10.3f}')


if __name__ == '__main__':
    main()



from math import sqrt, pi

import numpy as np
from ase.units import Bohr

from gpaw.utilities import pack
from gpaw.analyse.hirshfeld import HirshfeldDensity
from gpaw.utilities.tools import coordinates


def wignerseitz(gd, atoms, scale=None):
    """Determine which atom is closest to each grid point.

    The atom distances might be scaled by the scale list."""
    if scale is None:
        scale = [1.] * len(atoms)
    else:
        assert len(scale) == len(atoms)
    r_vG, R2min_G = coordinates(gd, atoms[0].position / Bohr)
    R2min_G *= scale[0] ** 2
    index_G = gd.zeros(dtype=int)
    for a, atom in enumerate(atoms[1:]):
        r_vG, r2_G = coordinates(gd, atom.position / Bohr)
        r2_G *= scale[a + 1] ** 2
        index_G = np.where(R2min_G > r2_G, a + 1, index_G)
        R2min_G = np.where(R2min_G > r2_G, r2_G, R2min_G)
    return index_G


class WignerSeitz:

    def __init__(self, gd, atoms,
                 calculator=None, scale=None):
        """Find the grid points nearest to the atoms"""

        self.atoms = atoms
        self.gd = gd
        self.calculator = calculator

        self.atom_index = wignerseitz(gd, atoms, scale)

    def expand(self, density):
        """Expand a smooth density in Wigner-Seitz cells around the atoms"""
        n = len(self.atoms)
        weights = np.empty((n,))
        for a in range(n):
            mask = np.where(self.atom_index == a, density, 0.0)
            # XXX Optimize! No need to integrate in zero-region
            weights[a] = self.gd.integrate(mask)

        return weights

    def expand_density(self, nt_G, s, nspins):
        """Get the weights of spin-density in Wigner-Seitz cells
        around the atoms. The spin index and number of spins are
        needed for the augmentation sphere corrections."""
        weights_a = self.expand(nt_G)
        for a, nucleus in enumerate(self.atoms):
            weights_a[a] += nucleus.get_density_correction(s, nspins)
        return weights_a

    def expand_wave_function(self, psit_G, u, n):
        """Get the weights of wave function in Wigner-Seitz cells
        around the atoms. The spin-k-point index u and band number n
        are needed for the augmentation sphere corrections."""

        assert psit_G.dtype == float
        # smooth part
        weigths = self.expand(psit_G ** 2)

        # add augmentation sphere corrections
        for a, nucleus in enumerate(self.atoms):
            P_i = nucleus.P_uni[u, n]
            P_p = pack(np.outer(P_i, P_i))
            Delta_p = sqrt(4 * pi) * nucleus.setup.Delta_pL[:, 0]
            weigths[a] += np.dot(Delta_p, P_p)

        return weigths

    def get_charges(self, den_g):
        """Charge on the atom according to the Wigner-Seitz partitioning

        Can be applied to any density den_g.
        """
        assert den_g.shape == tuple(self.gd.n_c)
        charges = []
        for atom, q in zip(self.atoms, self.expand(den_g)):
            charges.append(atom.number - q)
        return charges

    def get_effective_volume_ratio(self, atom_index):
        """Effective volume to free volume ratio.

        After: Tkatchenko and Scheffler PRL 102 (2009) 073005
        """
        atoms = self.atoms
        finegd = self.gd

        den_g, gd = self.calculator.density.get_all_electron_density(atoms)
        assert gd == finegd
        denfree_g, gd = self.hdensity.get_density([atom_index])
        assert gd == finegd

        # the atoms r^3 grid
        position = self.atoms[atom_index].position / Bohr
        r_vg, r2_g = coordinates(finegd, origin=position)
        r3_g = r2_g * np.sqrt(r2_g)

        weight_g = np.where(self.atom_index == atom_index, 1.0, 0.0)

        nom = finegd.integrate(r3_g * den_g[0] * weight_g)
        denom = finegd.integrate(r3_g * denfree_g)

        return nom / denom

    def get_effective_volume_ratios(self):
        """Return the list of effective volume to free volume ratios."""
        ratios = []
        self.hdensity = HirshfeldDensity(self.calculator)
        for a, atom in enumerate(self.atoms):
            ratios.append(self.get_effective_volume_ratio(a))
        return np.array(ratios)


class LDOSbyBand:

    """Base class for a band by band LDOS"""

    def by_element(self):
        # get element indicees
        elemi = {}
        for i, nucleus in enumerate(self.paw.atoms):
            symbol = nucleus.setup.symbol
            if symbol in elemi:
                elemi[symbol].append(i)
            else:
                elemi[symbol] = [i]
        for key in elemi.keys():
            elemi[key] = self.get(elemi[key])
        return elemi


'''
class WignerSeitzLDOS(LDOSbyBand):

    """Class to get the unfolded LDOS defined by Wigner-Seitz cells"""

    def __init__(self, paw):
        self.paw = paw
        self.ws = WignerSeitz(paw.gd, paw.atoms)

        nu = paw.nkpts * paw.nspins
        ldos = np.empty((nu, paw.nbands, len(paw.atoms)))
        for u, kpt in enumerate(paw.kpt_u):
            for n, psit_G in enumerate(kpt.psit_nG):
                # XXX variable ws undefined.  Probably self.ws
                ldos[u, n, :] = ws.expand_wave_function(psit_G, u, n)

    def write(self, filename, slavewrite=False):
        if self.world.rank == MASTER or slavewrite:
            paw = self.paw
            f = open(filename, 'w')

            nn = len(paw.atoms)

            for k in range(paw.nkpts):
                for s in range(paw.nspins):
                    u = s * paw.nkpts + k
                    for n in range(paw.nbands):
                        # avery: Added dummy loop body to make compiling work.
                        1
'''



from typing import Sequence, Union

import numpy as np

from gpaw.atom.basis import BasisMaker
from gpaw.atom.basis import QuasiGaussian
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.atom.configurations import parameters, parameters_extra
from gpaw.basis_data import BasisFunction
from gpaw.basis_data import parse_basis_name

# Module for generating basis sets that compose of usual basis sets
# augmented with Gaussian type orbital (GTO).
#
# GTOs are truncated and represented numerically.


def create_GTO_dictionary(l: Union[int, str], exponent: float):
    """Dictionary representing Gaussian type orbital.

    Parameters
    ----------
    l
        Angular momentum
    exponent
        Gaussian exponent
    """
    return create_CGTO_dictionary(l, [exponent], [1.0])


def create_CGTO_dictionary(l: Union[int, str],
                           exponents: Sequence[float],
                           coefficients: Sequence[float]):
    """Dictionary representing contracted Gaussian type orbital.

    Parameters
    ----------
    l
        Angular momentum
    exponents
        Gaussian exponents
    coefficients
        Gaussian coefficients
    """
    if isinstance(l, str):
        l = 'spdfghi'.index(l.lower())
    gto = {'angular_momentum': [l],
           'exponents': exponents,
           'coefficients': [coefficients]}
    return gto


def read_gaussian_basis_file(fname):
    """Read Gaussian basis set file.

    This reads only the first element/atom from the file
    as separated with line beginning with '*'.
    """
    gtos = []
    description = ''

    with open(fname, 'r') as fd:
        line_i = fd.readlines()

    i = 0
    Ni = len(line_i)
    while True:
        line = line_i[i].strip()
        if line == '' or line[0] == '*':
            pass
        elif line[0] == '!':
            description += '%s\n' % line[1:].strip()
        else:
            break
        i += 1
    description = description.strip()

    atom = line_i[i].strip().split()[0]
    i += 1
    while i < Ni:
        line = line_i[i]
        if line[0] == '*':
            break
        i += 1
        d = line.split()
        l = 'spdfghi'.index(d[0].lower())
        Nj = int(d[1])
        alpha_j = []
        coeff_j = []
        for _ in range(Nj):
            line = line_i[i]
            d = line.split()
            alpha = float(d[0].replace('D', 'E'))
            coeff = float(d[1].replace('D', 'E'))
            alpha_j.append(alpha)
            coeff_j.append(coeff)
            i += 1
        gto = create_CGTO_dictionary(l, alpha_j, coeff_j)
        gtos.append(gto)

    return atom, description, gtos


def get_ngto(rgd, l, alpha, rcut):
    gaussian = QuasiGaussian(alpha, rcut)
    psi_g = gaussian(rgd.r_g) * rgd.r_g**l
    norm = np.sum(rgd.dr_g * (rgd.r_g * psi_g)**2)**.5
    psi_g /= norm
    return psi_g


def create_ngto(rgd, l, alpha, rmax, tol):
    # Get NGTO with the initial (large) rcut=rmax
    psiref_g = get_ngto(rgd, l, alpha, rmax)

    # Make rcut smaller

    # Guess initial rcut where we are close to the tolerance
    i = np.where(psiref_g > tol)[0][-1]
    rcut = rgd.r_g[i]
    psi_g = get_ngto(rgd, l, alpha, rcut)
    err = np.max(np.absolute(psi_g - psiref_g))

    # Increase/decrease rcut to find the smallest rcut
    # that yields error within the tolerance
    if err > tol:
        # Increase rcut -> decrease err
        for i in range(i, len(rgd.r_g)):
            rcut = rgd.r_g[i]
            psi_g = get_ngto(rgd, l, alpha, rcut)
            err = np.max(np.absolute(psi_g - psiref_g))
            if err < tol:
                break
    else:
        # Decrease rcut -> increase err
        for i in range(i, 0, -1):
            rcut = rgd.r_g[i]
            psi_g = get_ngto(rgd, l, alpha, rcut)
            err = np.max(np.absolute(psi_g - psiref_g))
            if err > tol:
                i += 1
                break

    # Construct NGTO with the found rcut
    rcut = rgd.r_g[i]
    psi_g = get_ngto(rgd, l, alpha, rcut)

    # Change norm (maybe unnecessary)
    psi_g = psi_g[:(i + 1)] * 0.5

    return psi_g


def add_ngto(basis, l, coeff_j, alpha_j, tol, label):
    rgd = basis.get_grid_descriptor()
    rmax = rgd.r_g[-1]

    # Create linear combination of NGTO's
    psi_g = np.zeros(rgd.r_g.shape)
    i_max = 0
    for coeff, alpha in zip(coeff_j, alpha_j):
        contrib = coeff * create_ngto(rgd, l, alpha, rmax, tol)
        i = contrib.size
        i_max = max(i, i_max)
        psi_g[0:i] += contrib

    psi_g = psi_g[0:i_max]
    rcut = rgd.r_g[i_max]

    # Create associated basis function
    bf = BasisFunction(None, l, rcut, psi_g, label)
    basis.bf_j.append(bf)


def generate_nao_ngto_basis(atom, *, xc, nao, name,
                            gtos, gto_description=None,
                            rmax=100.0, tol=0.001):
    # Choose basis sets without semi-core states XXXXXX
    if atom == 'Ag':
        name = '11.%s' % name
        p = parameters_extra
    else:
        p = parameters

    # Generate nao basis
    zetacount, polarizationcount = parse_basis_name(nao)
    bm = BasisMaker(atom, name=name, run=False, gtxt=None, xc=xc)
    bm.generator.run(write_xml=False, **p[atom])
    basis = bm.generate(zetacount, polarizationcount, txt=None)

    # Increase basis function max radius
    assert isinstance(basis.rgd, EquidistantRadialGridDescriptor)
    h = basis.rgd.dr_g[0]
    assert basis.rgd.r_g[0] == 0.0
    N = int(rmax / h) + 1
    basis.rgd = EquidistantRadialGridDescriptor(h, N)

    # Add NGTOs
    description = []
    msg = 'Augmented with NGTOs'
    description.append(msg)
    description.append('=' * len(msg))
    description.append('')
    if gto_description is not None:
        description.append(gto_description)
        description.append('')
    description.append('NGTO truncation tolerance: %f' % tol)
    description.append('Functions: NGTO(l,coeff*alpha + ...)')

    for gto in gtos:
        assert len(gto['angular_momentum']) == 1
        l = gto['angular_momentum'][0]
        alpha_j = gto['exponents']
        # Float conversion
        alpha_j = [float(a) for a in alpha_j]
        for coeff_j in gto['coefficients']:
            assert len(alpha_j) == len(coeff_j)
            # Float conversion
            coeff_j = [float(c) for c in coeff_j]
            coeff_alpha_list = ['%+.3f*%.3f' % (c, a)
                                for c, a in zip(coeff_j, alpha_j)]
            coeff_alpha_label = ''.join(coeff_alpha_list[0:3])
            if len(coeff_alpha_list) > 3:
                coeff_alpha_label += '+...'
            ngtolabel = 'NGTO(%s,%s)' % ('spdfghi'[l], coeff_alpha_label)
            description.append('    ' + ngtolabel)
            add_ngto(basis, l, coeff_j, alpha_j, tol, ngtolabel)

    basis.generatordata += '\n\n' + '\n'.join(description)

    basis.write_xml()



from time import ctime
from functools import partial

import numpy as np
from ase.units import Ha

from gpaw.response.integrators import Integrand
from gpaw.response.chi0 import Chi0Calculator
from gpaw.response.pair_functions import SingleQPWDescriptor
from gpaw.response.chi0_data import Chi0DrudeData
from gpaw.response.frequencies import FrequencyGridDescriptor


class Chi0DrudeCalculator(Chi0Calculator):
    """Class for calculating the plasma frequency contribution to Chi0,
    that is, the contribution from intraband transitions inside of metallic
    bands. This corresponds directly to the dielectric function in the Drude
    model."""

    def __init__(self, pair,
                 disable_point_group=False,
                 disable_time_reversal=False,
                 disable_non_symmorphic=True,
                 integrationmode=None):
        self.pair = pair
        self.gs = pair.gs
        self.context = pair.context

        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal
        self.disable_non_symmorphic = disable_non_symmorphic
        self.integrationmode = integrationmode

        # Number of completely filled bands and number of non-empty bands.
        self.nocc1, self.nocc2 = self.gs.count_occupied_bands()

    def calculate(self, wd, rate, spin='all'):
        """Calculate the Drude dielectric response.

        Parameters
        ----------
        wd : FrequencyDescriptor
            Frequencies to evaluate the reponse function at.
        rate : float
            Plasma frequency decay rate (in eV), corresponding to the
            imaginary part of the complex frequency.
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.
        """
        self.print_info(wd, rate)

        # Parse the spin input
        spins = self.get_spins(spin)

        chi0_drude = Chi0DrudeData.from_frequency_descriptor(wd, rate)
        self._calculate(chi0_drude, spins)

        return chi0_drude

    def _calculate(self, chi0_drude: Chi0DrudeData, spins):
        """In-place calculation of the Drude dielectric response function,
        based on the free-space plasma frequency of the intraband transitions.
        """
        # Create a dummy plane-wave descriptor. We need this for the symmetry
        # analysis -> see discussion in gpaw.response.jdos
        qpd = SingleQPWDescriptor.from_q([0., 0., 0.],
                                         ecut=1e-3, gd=self.gs.gd)

        integrator = self.initialize_integrator()
        domain, analyzer, prefactor = self.get_integration_domain(qpd, spins)
        kind, extraargs = self.get_integral_kind()

        integrand = PlasmaFrequencyIntegrand(self, qpd, analyzer)

        # Integrate using temporary array
        tmp_plasmafreq_wvv = np.zeros((1,) + chi0_drude.vv_shape, complex)
        integrator.integrate(kind=kind,  # Kind of integral
                             domain=domain,  # Integration domain
                             integrand=integrand,
                             out_wxx=tmp_plasmafreq_wvv,  # Output array
                             **extraargs)  # Extra args for int. method
        tmp_plasmafreq_wvv *= prefactor

        # Store the plasma frequency itself and print it for anyone to use
        plasmafreq_vv = tmp_plasmafreq_wvv[0].copy()
        analyzer.symmetrize_wvv(plasmafreq_vv[np.newaxis])
        chi0_drude.plasmafreq_vv += 4 * np.pi * plasmafreq_vv
        self.context.print('Plasma frequency:', flush=False)
        self.context.print((chi0_drude.plasmafreq_vv**0.5 * Ha).round(2))

        # Calculate the Drude dielectric response function from the
        # free-space plasma frequency
        # χ_D(ω+iη) = ω_p^2 / (ω+iη)^2
        assert chi0_drude.zd.upper_half_plane
        chi0_drude.chi_Zvv += plasmafreq_vv[np.newaxis] \
            / chi0_drude.zd.hz_z[:, np.newaxis, np.newaxis]**2

    def update_integrator_kwargs(self, *unused, **ignored):
        """The Drude calculator uses only standard integrator kwargs."""
        pass

    def get_integral_kind(self):
        """Define what "kind" of integral to make."""
        extraargs = {}
        # The plasma frequency integral is special in the way, that only
        # the spectral part is needed
        kind = 'spectral function'
        if self.integrationmode is None:
            # Calculate intraband transitions at finite fermi smearing
            extraargs['intraband'] = True  # Calculate intraband
        elif self.integrationmode == 'tetrahedron integration':
            # Calculate intraband transitions at T=0
            fermi_level = self.gs.fermi_level
            extraargs['x'] = FrequencyGridDescriptor([-fermi_level])

        return kind, extraargs

    def print_info(self, wd, rate):
        p = partial(self.context.print, flush=False)

        p('%s' % ctime())
        p('Called chi0_drude.calculate() with:')
        p('    Number of frequency points: %d' % len(wd))
        p('    Plasma frequency decay rate: %f eV' % rate)
        p()
        p(self.get_gs_info_string(tab='    '))
        self.context.print('')


class PlasmaFrequencyIntegrand(Integrand):
    def __init__(self, chi0drudecalc, qpd, analyzer):
        self._drude = chi0drudecalc
        self.qpd = qpd
        self.analyzer = analyzer

    def _band_summation(self):
        # Intraband response needs only integrate partially unoccupied bands.
        return self._drude.nocc1, self._drude.nocc2

    def matrix_element(self, k_v, s):
        """NB: In dire need of documentation! XXX."""
        n1, n2 = self._band_summation()
        k_c = np.dot(self.qpd.gd.cell_cv, k_v) / (2 * np.pi)
        kpt1 = self._drude.pair.get_k_point(s, k_c, n1, n2)
        n_n = range(n1, n2)

        vel_nv = self._drude.pair.intraband_pair_density(kpt1, n_n)

        if self._drude.integrationmode is None:
            f_n = kpt1.f_n
            width = self._drude.gs.get_occupations_width()
            if width > 1e-15:
                dfde_n = - 1. / width * (f_n - f_n**2.0)
            else:
                dfde_n = np.zeros_like(f_n)
            vel_nv *= np.sqrt(-dfde_n[:, np.newaxis])
            weight = np.sqrt(self.analyzer.get_kpoint_weight(k_c) /
                             self.analyzer.how_many_symmetries())
            vel_nv *= weight

        return vel_nv

    def eigenvalues(self, k_v, s):
        """A function that can return the intraband eigenvalues.

        A method describing the integrand of
        the response function which gives an output that
        is compatible with the gpaw k-point integration
        routines."""
        n1, n2 = self._band_summation()
        gs = self._drude.gs
        kd = gs.kd
        k_c = np.dot(self.qpd.gd.cell_cv, k_v) / (2 * np.pi)
        K1 = self._drude.pair.find_kpoint(k_c)
        ik = kd.bz2ibz_k[K1]
        kpt1 = gs.kpt_qs[ik][s]
        assert gs.kd.comm.size == 1

        return kpt1.eps_n[n1:n2]



"""Calculate the Heisenberg exchange constants in Fe and Co using the MFT.
Test with unrealisticly loose parameters to catch if the numerics change.
"""

# General modules
import pytest
import numpy as np

# Script modules
from gpaw import GPAW

from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.localft import LocalFTCalculator, LocalPAWFTCalculator
from gpaw.response.mft import IsotropicExchangeCalculator
from gpaw.response.site_kernels import (SphericalSiteKernels,
                                        CylindricalSiteKernels,
                                        ParallelepipedicSiteKernels)
from gpaw.response.heisenberg import (calculate_single_site_magnon_energies,
                                      calculate_fm_magnon_energies)
from gpaw.test.conftest import response_band_cutoff


@pytest.mark.response
def test_Fe_bcc(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    # MFT calculation
    ecut = 50
    # Do the high symmetry points of the bcc lattice
    q_qc = np.array([[0, 0, 0],           # Gamma
                     [0.5, -0.5, 0.5],    # H
                     [0.0, 0.0, 0.5],     # N
                     ])
    # Define site kernels to test
    # Test a single site of spherical and cylindrical geometries
    rc_pa = np.array([[1.0], [1.5], [2.0]])
    hc_pa = np.array([[1.0], [1.5], [2.0]])
    ez_pav = np.array([[[1., 0., 0.]], [[0., 1., 0.]], [[0., 0., 1.]]])

    # ---------- Script ---------- #

    # Extract the ground state fixture
    calc = GPAW(gpw_files['fe_pw_wfs'], parallel=dict(domain=1))
    nbands = response_band_cutoff['fe_pw_wfs']
    atoms = calc.atoms

    # Set up site kernels with a single site
    positions = atoms.get_positions()
    sitekernels = SphericalSiteKernels(positions, rc_pa)
    sitekernels.append(CylindricalSiteKernels(positions, ez_pav,
                                              rc_pa, hc_pa))
    # Set up a kernel to fill out the entire unit cell
    sitekernels.append(ParallelepipedicSiteKernels(positions,
                                                   [[atoms.get_cell()]]))

    # Initialize the Heisenberg exchange calculator
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext()
    chiks_calc = ChiKSCalculator(gs, context,
                                 ecut=ecut, nbands=nbands, gammacentered=True)
    localft_calc = LocalFTCalculator.from_rshe_parameters(gs, context)
    isoexch_calc = IsotropicExchangeCalculator(chiks_calc, localft_calc)

    # Allocate array for the exchange constants
    nq = len(q_qc)
    nsites = sitekernels.nsites
    npartitions = sitekernels.npartitions
    J_qabp = np.empty((nq, nsites, nsites, npartitions), dtype=complex)

    # Calcualate the exchange constant for each q-point
    for q, q_c in enumerate(q_qc):
        J_qabp[q] = isoexch_calc(q_c, sitekernels)

    # Since we only have a single site, reduce the array
    J_qp = J_qabp[:, 0, 0, :]

    # Calculate the magnon energies
    mm = 2.21
    mm_ap = mm * np.ones((1, npartitions))  # Magnetic moments
    mw_qp = calculate_fm_magnon_energies(J_qabp, q_qc, mm_ap)[:, 0, :]

    # Compare results to test values
    test_J_pq = np.array(
        [[2.1907596825086455, 1.172424411323134, 1.6060583789867644],
         [2.612428039019977, 1.2193926800088601, 1.7635196888465006],
         [6.782367391186284, 0.2993922109834177, 1.9346016211386057],
         [1.5764800860123762, 0.8365204592352894, 1.1648584638500161],
         [2.4230224513213234, 1.2179759558303274, 1.6691805687218078],
         [5.35668502504496, 0.3801778545994659, 1.6948968244858478],
         [2.523580017606111, 1.21779750159267, 1.7637120466695273]])
    test_mw_pq = np.array(
        [[0.0, 0.9215703811633589, 0.5291414511510236],
         [0.0, 1.2606654832679791, 0.7682428508357253],
         [0.0, 5.866945864436984, 4.38711834393455],
         [0.0, 0.6696467210652369, 0.3725082553505521],
         [0.0, 1.0905398149239784, 0.682209848506349],
         [0.0, 4.503626398593207, 3.313835475619106],
         [0.0, 1.181703634401304, 0.6876633221145555]])

    # Exchange constants
    assert J_qp.imag == pytest.approx(0.0)
    assert J_qp.T.real == pytest.approx(test_J_pq, rel=2e-3)

    # Magnon energies
    assert mw_qp.T == pytest.approx(test_mw_pq, rel=2e-3)


@pytest.mark.response
def test_Co_hcp(in_tmp_dir, gpw_files):
    # ---------- Inputs ---------- #

    # MFT calculation
    ecut = 100
    # Do high symmetry points of the hcp lattice
    q_qc = np.array([[0, 0, 0],              # Gamma
                     [0.5, 0., 0.],          # M
                     [0., 0., 0.5]           # A
                     ])

    # Use spherical site kernels in a radius range which should yield
    # stable results
    rc_pa = np.array([[1.0, 1.0], [1.1, 1.1], [1.2, 1.2]])

    # Unfortunately, the usage of symmetry leads to such extensive repetition
    # of random noise, that one cannot trust individual values of J very well.
    # This is improved when increasing the number of k-points, but the problem
    # never completely vanishes
    J_atol = 5.e-3
    J_rtol = 5.e-2
    # However, derived physical values have an increased error cancellation due
    # to their collective nature.
    mw_rtol = 5.e-3  # relative tolerance of absolute results
    mw_ctol = 5.e-2  # relative tolerance on kernel and eta self-consistency

    # ---------- Script ---------- #

    # Extract the ground state fixture
    calc = GPAW(gpw_files['co_pw_wfs'], parallel=dict(domain=1))
    nbands = response_band_cutoff['co_pw_wfs']
    atoms = calc.get_atoms()

    # Set up spherical site kernels
    positions = atoms.get_positions()
    sitekernels = SphericalSiteKernels(positions, rc_pa)

    # Set up a site kernel to fill out the entire unit cell
    cell_cv = atoms.get_cell()
    cc_v = np.sum(cell_cv, axis=0) / 2.  # Unit cell center
    ucsitekernels = ParallelepipedicSiteKernels([cc_v], [[cell_cv]])

    # Initialize the exchange calculator with and without symmetry
    gs = ResponseGroundStateAdapter(calc)
    context = ResponseContext()
    chiks_calc0 = ChiKSCalculator(gs, context,
                                  disable_point_group=True,
                                  disable_time_reversal=True,
                                  ecut=ecut, nbands=nbands, gammacentered=True)
    localft_calc = LocalPAWFTCalculator(gs, context)
    isoexch_calc0 = IsotropicExchangeCalculator(chiks_calc0, localft_calc)
    chiks_calc1 = ChiKSCalculator(gs, context,
                                  ecut=ecut, nbands=nbands, gammacentered=True)
    isoexch_calc1 = IsotropicExchangeCalculator(chiks_calc1, localft_calc)

    # Allocate array for the spherical site exchange constants
    nq = len(q_qc)
    nsites = sitekernels.nsites
    npartitions = sitekernels.npartitions
    J_qabp = np.empty((nq, nsites, nsites, npartitions), dtype=complex)

    # Allocate array for the unit cell site exchange constants
    Juc_qs = np.empty((nq, 2), dtype=complex)

    # Calcualate the exchange constants for each q-point
    for q, q_c in enumerate(q_qc):
        J_qabp[q] = isoexch_calc0(q_c, sitekernels)
        chiksr_buffer = isoexch_calc0._chiksr
        Juc_qs[q, 0] = isoexch_calc0(q_c, ucsitekernels)[0, 0, 0]
        assert isoexch_calc0._chiksr is chiksr_buffer,\
            'Two subsequent IsotropicExchangeCalculator calls with the same '\
            'q_c, should reuse, not update, the chiks buffer'

        Juc_qs[q, 1] = isoexch_calc1(q_c, ucsitekernels)[0, 0, 0]

    # Calculate the magnon energy
    mom = atoms.get_magnetic_moment()
    mm_ap = mom / 2.0 * np.ones((nsites, npartitions))
    mw_qnp = calculate_fm_magnon_energies(J_qabp, q_qc, mm_ap)
    mw_qnp = np.sort(mw_qnp, axis=1)  # Make sure the eigenvalues are sorted
    mwuc_qs = calculate_single_site_magnon_energies(Juc_qs, q_qc, mom)

    # Compare results to test values
    # print(J_qabp[..., 1])
    # print(mw_qnp[..., 1])
    # print(mwuc_qs[:, 0])
    test_J_qab = np.array([[[1.23106207 - 0.j, 0.25816335 - 0.j],
                            [0.25816335 + 0.j, 1.23106207 + 0.j]],
                           [[0.88823839 + 0.j, 0.07345416 - 0.04947835j],
                            [0.07345416 + 0.04947835j, 0.88823839 + 0.j]],
                           [[1.09349955 - 0.j, 0.00000010 - 0.01176761j],
                            [0.00000010 + 0.01176761j, 1.09349955 - 0.j]]])
    test_mw_qn = np.array([[0., 0.64793939],
                           [0.64304039, 0.86531921],
                           [0.48182997, 0.51136436]])
    test_mwuc_q = np.array([0., 0.69678659, 0.44825874])

    # Exchange constants
    # err = np.absolute(J_qabp[..., 1] - test_J_qab)
    # is_bad = err > J_atol + J_rtol * np.absolute(test_J_qab)
    # print(is_bad)
    # print(np.absolute(err[is_bad] / np.absolute(test_J_qab[is_bad])))
    assert np.allclose(J_qabp[..., 1], test_J_qab,
                       atol=J_atol, rtol=J_rtol)

    # Magnon energies
    assert np.all(np.abs(mw_qnp[0, 0, :]) < 1.e-8)  # Goldstone theorem
    assert np.allclose(mwuc_qs[0, :], 0.)  # Goldstone
    assert np.allclose(mw_qnp[1:, 0, 1], test_mw_qn[1:, 0], rtol=mw_rtol)
    assert np.allclose(mw_qnp[:, 1, 1], test_mw_qn[:, 1], rtol=mw_rtol)
    assert np.allclose(mwuc_qs[1:, 0], test_mwuc_q[1:], rtol=mw_rtol)

    # Check self-consistency of results
    # We should be in a radius range, where the magnon energies don't change
    assert np.allclose(mw_qnp[1:, 0, ::2],
                       test_mw_qn[1:, 0, np.newaxis], rtol=mw_ctol)
    assert np.allclose(mw_qnp[:, 1, ::2],
                       test_mw_qn[:, 1, np.newaxis], rtol=mw_ctol)
    # Check that symmetry toggle do not change the magnon energies
    assert np.allclose(mwuc_qs[1:, 0], mwuc_qs[1:, 1], rtol=mw_ctol)



import numpy as np
import pytest

from ase.build import molecule
from ase.utils import workdir

from gpaw import GPAW
from gpaw.mpi import world, serial_comm, broadcast
from gpaw.lcaotddft.wfwriter import WaveFunctionReader
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.frequencydensitymatrix import FrequencyDensityMatrix
from gpaw.lcaotddft.ksdecomposition import KohnShamDecomposition

from gpaw.test import only_on_master
from . import (parallel_options, calculate_time_propagation, calculate_error,
               check_txt_data, check_wfs)

pytestmark = pytest.mark.usefixtures('module_tmp_path')

parallel_i = parallel_options(fix_sl_auto=True)


def calculate_ground_state(*, communicator=world,
                           spinpol=False):
    atoms = molecule('NaCl')
    atoms.center(vacuum=4.0)
    calc = GPAW(nbands=6,
                h=0.4,
                setups=dict(Na='1'),
                basis='dzp',
                mode='lcao',
                convergence={'density': 1e-8},
                spinpol=spinpol,
                communicator=communicator,
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')
    return calc


@pytest.fixture(scope='module')
@only_on_master(world)
def initialize_system():
    comm = serial_comm

    calc = calculate_ground_state(communicator=comm)
    fdm = calculate_time_propagation('gs.gpw',
                                     kick=np.ones(3) * 1e-5,
                                     communicator=comm,
                                     do_fdm=True)

    # Calculate ground state with full unoccupied space
    unocc_calc = calc.fixed_density(nbands='nao',
                                    communicator=comm,
                                    txt='unocc.out')
    unocc_calc.write('unocc.gpw', mode='all')
    return unocc_calc, fdm


@pytest.mark.rttddft
def test_propagated_wave_function(initialize_system, module_tmp_path):
    wfr = WaveFunctionReader(module_tmp_path / 'wf.ulm')
    coeff = wfr[-1].wave_functions.coefficients
    # Pick a few coefficients corresponding to non-degenerate states;
    # degenerate states should be normalized so that they can be compared
    coeff = coeff[np.ix_([0], [0], [0, 1, 4], [0, 1, 2])]
    # Normalize the wave function sign
    coeff = np.sign(coeff.real[..., 0, np.newaxis]) * coeff
    ref = [[[[1.6564776755628504e-02 + 1.2158943340143986e-01j,
              4.7464497657284752e-03 + 3.4917799444496286e-02j,
              8.2152048273399657e-07 - 1.6344333784831069e-06j],
             [1.5177089239371724e-01 + 7.6502712023931621e-02j,
              8.0497556154952932e-01 + 4.0573839188792121e-01j,
              -5.1505952970811632e-06 - 1.1507918955641119e-05j],
             [2.5116252101774323e+00 + 3.6776360873471503e-01j,
              1.9024613198566329e-01 + 2.7843314959952882e-02j,
              -1.3848736953929574e-05 - 2.6402210145403184e-05j]]]]
    err = calculate_error(coeff, ref)
    assert err < 3e-12


@pytest.mark.rttddft
@pytest.mark.parametrize('parallel', parallel_i)
def test_propagation(initialize_system, module_tmp_path, parallel, in_tmp_dir):
    calculate_time_propagation(module_tmp_path / 'gs.gpw',
                               kick=np.ones(3) * 1e-5,
                               parallel=parallel)
    check_wfs(module_tmp_path / 'wf.ulm', 'wf.ulm', atol=1e-12)


@pytest.fixture(scope='module')
@only_on_master(world, broadcast=broadcast)
def dipole_moment_reference(initialize_system):
    from gpaw.tddft.spectrum import \
        read_dipole_moment_file, calculate_fourier_transform

    unocc_calc, fdm = initialize_system
    _, time_t, _, dm_tv = read_dipole_moment_file('dm.dat')
    dm_tv = dm_tv - dm_tv[0]
    dm_wv = calculate_fourier_transform(time_t, dm_tv,
                                        fdm.foldedfreqs_f[0])
    return dm_wv


@pytest.fixture(scope='module')
@only_on_master(world)
def ksd_reference(initialize_system):
    unocc_calc, fdm = initialize_system
    ksd = KohnShamDecomposition(unocc_calc)
    ksd.initialize(unocc_calc)
    return ksd, fdm


def ksd_transform_fdm(ksd, fdm):
    rho_iwp = np.empty((2, len(fdm.freq_w), len(ksd.w_p)), dtype=complex)
    rho_iwp[:] = np.nan + 1j * np.nan
    for i, rho_wuMM in enumerate([fdm.FReDrho_wuMM, fdm.FImDrho_wuMM]):
        for w in range(len(fdm.freq_w)):
            rho_uMM = rho_wuMM[w]
            rho_up = ksd.transform(rho_uMM)
            rho_iwp[i, w, :] = rho_up[0]
    return rho_iwp


@pytest.fixture(scope='module')
@only_on_master(world, broadcast=broadcast)
def ksd_transform_reference(ksd_reference):
    ksd, fdm = ksd_reference
    ref_rho_iwp = ksd_transform_fdm(ksd, fdm)
    return ref_rho_iwp


@pytest.fixture(scope='module', params=parallel_i)
def build_ksd(initialize_system, request):
    calc = GPAW('unocc.gpw', parallel=request.param, txt=None)
    if not calc.wfs.ksl.using_blacs and calc.wfs.bd.comm.size > 1:
        pytest.xfail('Band parallelization without scalapack '
                     'is not supported')
    ksd = KohnShamDecomposition(calc)
    ksd.initialize(calc)
    ksd.write('ksd.ulm')


@pytest.fixture(scope='module', params=parallel_i)
def load_ksd(build_ksd, request):
    calc = GPAW('unocc.gpw', parallel=request.param, txt=None)
    # Initialize positions in order to calculate density
    calc.initialize_positions()
    ksd = KohnShamDecomposition(calc, 'ksd.ulm')
    dmat = DensityMatrix(calc)
    fdm = FrequencyDensityMatrix(calc, dmat, 'fdm.ulm')
    return ksd, fdm


@pytest.fixture(scope='module')
def ksd_transform(load_ksd):
    ksd, fdm = load_ksd
    rho_iwp = ksd_transform_fdm(ksd, fdm)
    return rho_iwp


@pytest.mark.rttddft
def test_ksd_transform(ksd_transform, ksd_transform_reference):
    ref_iwp = ksd_transform_reference
    rho_iwp = ksd_transform
    err = calculate_error(rho_iwp, ref_iwp)
    atol = 1e-18
    assert err < atol


@pytest.mark.rttddft
def test_ksd_transform_real_only(load_ksd, ksd_transform_reference):
    ksd, fdm = load_ksd
    ref_iwp = ksd_transform_reference
    rho_iwp = np.empty((2, len(fdm.freq_w), len(ksd.w_p)), dtype=complex)
    rho_iwp[:] = np.nan + 1j * np.nan
    for i, rho_wuMM in enumerate([fdm.FReDrho_wuMM, fdm.FImDrho_wuMM]):
        for w in range(len(fdm.freq_w)):
            rho_uMM = rho_wuMM[w]
            rho_p = ksd.transform([rho_uMM[0].real], broadcast=True)[0] \
                + 1j * ksd.transform([rho_uMM[0].imag], broadcast=True)[0]
            rho_iwp[i, w, :] = rho_p
    err = calculate_error(rho_iwp, ref_iwp)
    atol = 1e-18
    assert err < atol


@pytest.mark.rttddft
def test_dipole_moment_from_ksd(ksd_transform, load_ksd,
                                dipole_moment_reference):
    ksd, fdm = load_ksd
    dm_wv = np.empty((len(fdm.freq_w), 3), dtype=complex)
    dm_wv[:] = np.nan + 1j * np.nan
    rho_wp = ksd_transform[0]
    for w in range(len(fdm.freq_w)):
        dm_v = ksd.get_dipole_moment([rho_wp[w]])
        dm_wv[w, :] = dm_v

    ref_wv = dipole_moment_reference
    err = calculate_error(dm_wv, ref_wv)
    atol = 1e-7
    assert err < atol


def get_density_fdm(ksd, fdm, kind):
    assert kind in ['dmat', 'ksd']
    rho_wg = fdm.dmat.density.finegd.empty(len(fdm.freq_w), dtype=complex)
    rho_wg[:] = np.nan + 1j * np.nan
    for w in range(len(fdm.freq_w)):
        rho_uMM = fdm.FReDrho_wuMM[w]
        if kind == 'dmat':
            rho_g = fdm.dmat.get_density([rho_uMM[0].real]) \
                + 1j * fdm.dmat.get_density([rho_uMM[0].imag])
        elif kind == 'ksd':
            rho_up = ksd.transform(rho_uMM, broadcast=True)
            rho_g = ksd.get_density(fdm.dmat.wfs, [rho_up[0].real]) \
                + 1j * ksd.get_density(fdm.dmat.wfs, [rho_up[0].imag])
        rho_wg[w, :] = rho_g
    return rho_wg


@pytest.fixture(scope='module')
@only_on_master(world, broadcast=broadcast)
def density_reference(ksd_reference):
    ksd, fdm = ksd_reference
    dmat_rho_wg = get_density_fdm(ksd, fdm, 'dmat')
    ksd_rho_wg = get_density_fdm(ksd, fdm, 'ksd')
    return dict(dmat=dmat_rho_wg, ksd=ksd_rho_wg)


@pytest.mark.rttddft
def test_ksd_vs_dmat_density(density_reference):
    ref_wg = density_reference['dmat']
    rho_wg = density_reference['ksd']
    err = calculate_error(rho_wg, ref_wg)
    atol = 2e-10
    assert err < atol


@pytest.fixture(scope='module')
def density(load_ksd):
    ksd, fdm = load_ksd
    if ksd.ksl.using_blacs:
        pytest.xfail('Scalapack is not supported')
    dmat_rho_wg = get_density_fdm(ksd, fdm, 'dmat')
    ksd_rho_wg = get_density_fdm(ksd, fdm, 'ksd')
    return dict(dmat=dmat_rho_wg, ksd=ksd_rho_wg)


@pytest.mark.rttddft
@pytest.mark.parametrize('kind', ['ksd', 'dmat'])
def test_density(kind, density, load_ksd, density_reference):
    ksd, fdm = load_ksd
    ref_wg = density_reference[kind]
    rho_wg = fdm.dmat.density.finegd.collect(density[kind])
    err = calculate_error(rho_wg, ref_wg)
    atol = 3e-19
    assert err < atol


@pytest.mark.rttddft
@pytest.mark.parametrize('kind', ['ksd', 'dmat'])
def test_dipole_moment_from_density(kind, density, load_ksd,
                                    dipole_moment_reference):
    ksd, fdm = load_ksd
    rho_wg = density[kind]
    dm_wv = np.empty((len(fdm.freq_w), 3), dtype=complex)
    dm_wv[:] = np.nan + 1j * np.nan
    for w in range(len(fdm.freq_w)):
        dm_v = ksd.density.finegd.calculate_dipole_moment(rho_wg[w])
        dm_wv[w, :] = dm_v

    ref_wv = dipole_moment_reference
    err = calculate_error(dm_wv, ref_wv)
    atol = 5e-7
    assert err < atol


@pytest.fixture(scope='module')
@only_on_master(world)
@workdir('spinpol', mkdir=True)
def initialize_system_spinpol():
    comm = serial_comm
    calculate_ground_state(communicator=comm, spinpol=True)
    calculate_time_propagation('gs.gpw',
                               kick=np.ones(3) * 1e-5,
                               communicator=comm,
                               do_fdm=True)


@pytest.mark.rttddft
def test_spinpol_dipole_moment(initialize_system, initialize_system_spinpol,
                               module_tmp_path):
    # The test system has even number of electrons and is non-magnetic
    # so spin-paired and spin-polarized calculation should give same result
    check_txt_data(module_tmp_path / 'dm.dat',
                   module_tmp_path / 'spinpol' / 'dm.dat',
                   atol=5e-14)


@pytest.mark.rttddft
@pytest.mark.parametrize('parallel', parallel_i)
def test_spinpol_propagation(initialize_system_spinpol, module_tmp_path,
                             parallel, in_tmp_dir):
    ref_path = module_tmp_path / 'spinpol'
    calculate_time_propagation(ref_path / 'gs.gpw',
                               kick=np.ones(3) * 1e-5,
                               parallel=parallel)
    check_wfs(ref_path / 'wf.ulm', 'wf.ulm', atol=1e-12)



"""Test automagical calculation of wfs"""

from gpaw import GPAW
from ase import Atoms

# H2


def test_fileio_wfs_auto(in_tmp_dir):
    H = Atoms('HH', [(0, 0, 0), (0, 0, 1)])
    H.center(vacuum=2.0)

    calc = GPAW(nbands=2,
                convergence={'eigenstates': 1e-3})
    H.calc = calc
    H.get_potential_energy()
    calc.write('tmp')

    calc = GPAW('tmp')
    calc.converge_wave_functions()



from ase import Atoms
from ase.data.vdw import vdw_radii
from gpaw.test import equal
from ase.units import Pascal, m
from gpaw.solvation import (
    SolvationGPAW,
    EffectivePotentialCavity,
    Power12Potential,
    LinearDielectric,
    KB51Volume,
    GradientSurface,
    VolumeInteraction,
    SurfaceInteraction,
    LeakedDensityInteraction)
import numpy as np

h = 0.2
d = 2.5
min_vac = 4.0
u0 = .180
epsinf = 80.
T = 298.15


def atomic_radii(atoms):
    return [vdw_radii[n] for n in atoms.numbers]


def test_solvation_forces_symmetry():
    xy_cell = np.ceil((min_vac * 2.) / h / 8.) * 8. * h
    z_cell = np.ceil((min_vac * 2. + d) / h / 8.) * 8. * h
    atoms = Atoms(
        'NaCl', positions=(
            (xy_cell / 2., xy_cell / 2., z_cell / 2. - d / 2.),
            (xy_cell / 2., xy_cell / 2., z_cell / 2. + d / 2.)
        )
    )
    atoms.set_cell((xy_cell, xy_cell, z_cell))

    atoms.calc = SolvationGPAW(
        xc='PBE', h=h, setups={'Na': '1'},
        cavity=EffectivePotentialCavity(
            effective_potential=Power12Potential(atomic_radii, u0),
            temperature=T,
            volume_calculator=KB51Volume(),
            surface_calculator=GradientSurface()
        ),
        dielectric=LinearDielectric(epsinf=epsinf),
        # parameters chosen to give ~ 1eV for each interaction
        interactions=[
            VolumeInteraction(pressure=-1e9 * Pascal),
            SurfaceInteraction(surface_tension=100. * 1e-3 * Pascal * m),
            LeakedDensityInteraction(voltage=10.)
        ]
    )
    F = atoms.calc.get_forces(atoms)

    difference = F[0][2] + F[1][2]
    print(difference)
    equal(difference, .0, .02)  # gas phase is ~.007 eV / Ang
    F[0][2] = F[1][2] = .0
    print(np.abs(F))
    equal(np.abs(F), .0, 1e-10)  # gas phase is ~1e-11 eV / Ang



from gpaw.gpu import cupy_is_fake
import _gpaw
try:
    from _gpaw_gpu import pwlfc_expand_gpu as pwacf_expand
except ImportError:
    def pwacf_expand(f_Gs, emiGR_Ga, Y_GL,
                     l_s, a_J, s_J,
                     cc, f_GI, I_J):
        assert cupy_is_fake
        _gpaw.pwlfc_expand(f_Gs._data, emiGR_Ga._data, Y_GL._data,
                           l_s._data, a_J._data, s_J._data,
                           cc, f_GI._data)



# General modules
import pytest
import numpy as np

# Script modules
from ase.build import bulk

from gpaw import GPAW, PW, FermiDirac
from gpaw.response.chi0 import Chi0


@pytest.mark.response
@pytest.mark.serial
def test_si_update_consistency(in_tmp_dir):
    """Test that we get consistent results, when calculating
    chi0 in one or multiple calls to update_chi0."""

    # ---------- Inputs ---------- #

    # Ground state calculation
    xc = 'LDA'
    a = 5.431
    pw = 200
    nbands = 8
    kpts = 4
    occw = 0.05

    # Response calculation
    q_c = np.array([0., 0., 0.])
    intermediate_m = 5

    # ---------- Script ---------- #
    
    # Ground state calculation
    atoms = bulk('Si', 'diamond', a=a)
    atoms.center()
    calc = GPAW(mode=PW(pw),
                nbands=nbands,
                kpts=(kpts, kpts, kpts),
                parallel={'domain': 1},
                occupations=FermiDirac(width=occw),
                xc=xc)

    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('Si', 'all')

    # Response calculation
    chi0 = Chi0('Si',
                hilbert=True,
                intraband=False)

    chi0_full = chi0.create_chi0(q_c)
    chi0_steps = chi0.create_chi0(q_c)
    spins = range(chi0.gs.nspins)
    # Add chi0 contribution from all the unoccupied bands
    chi0.update_chi0(chi0_full, chi0.nocc1, chi0.nbands, spins)
    # Add chi0 contribution from *some* of the unoccupied bands
    chi0.update_chi0(chi0_steps, chi0.nocc1, intermediate_m, spins)
    # Add chi0 contribution from the remaining unoccupied bands
    chi0.update_chi0(chi0_steps, intermediate_m, chi0.nbands, spins)

    # Compare the output chi0 body
    assert chi0_steps.chi0_WgG == pytest.approx(chi0_full.chi0_WgG,
                                                abs=1e-8, rel=1e-6)



from gpaw.atom.atompaw import AtomPAW
from gpaw.upf import UPFSetupData
import gpaw.test.pseudopotential.H_pz_hgh as H_hgh
import gpaw.test.pseudopotential.O_pz_hgh as O_hgh
from gpaw.mpi import world


def test_pseudopotential_atompaw(in_tmp_dir):
    def check(title, calc, epsref_n, threshold):
        eps_n = calc.wfs.kpt_u[0].eps_n
        print(title)
        for i, epsref in enumerate(epsref_n):
            err = abs(epsref - eps_n[i])
            ok = (err <= threshold)
            status = 'ok' if ok else 'FAILED'
            print('state %d | eps=%f | ref=%f | err=%f | tol=%s | %s' %
                  (i,
                   eps_n[i],
                   epsref,
                   err,
                   threshold,
                   status))
            assert ok

    # Load pseudopotential from Python module as string, then
    # write the string to a file, then load the file.
    def upf(upf_module, fname):
        with open(fname, 'w') as fd:
            fd.write(upf_module.ps_txt)
        return UPFSetupData(fname)

    if world.rank == 0:  # This test is not really parallel
        kwargs = dict(txt=None)
        for setup in ['paw', 'hgh', upf(H_hgh, 'H.pz-hgh.UPF')]:
            calc = AtomPAW('H', [[[1]]],
                           rcut=12.0, h=0.05,
                           setups={'H': setup}, **kwargs)
            # horrible UPF right now:
            tol = 5e-4 if setup in ['paw', 'hgh'] else 1e-3
            check('H %s' % setup, calc, [-0.233471], tol)

        for setup in ['paw', 'hgh', upf(O_hgh, 'O.pz-hgh.UPF')]:
            calc = AtomPAW('O', [[[2], [4]]],
                           rcut=10.0, h=0.035,
                           setups={'O': setup}, **kwargs)
            # horrible UPF right now:
            tol = 1e-3 if setup in ['paw', 'hgh'] else 5e-3
            check('O %s' % setup, calc, [-0.871362, -0.338381], tol)



import pytest
from ase.build import bulk
from gpaw import GPAW, LCAO

# This test calculates a GLLB quasiparticle gap with LCAO and verifies
# that it does not change from a reference value.  Note that the
# calculation, physically speaking, is garbage.


@pytest.fixture
def calc():
    with GPAW(mode=LCAO(interpolation=2),
              h=0.3,
              basis='sz(dzp)',
              xc='GLLBSC',
              kpts={'size': (2, 2, 2), 'gamma': True},
              convergence={'maximum iterations': 1},
              txt='si.txt') as calc:
        yield calc


@pytest.mark.later
@pytest.mark.gllb
@pytest.mark.libxc
def test_lcao_gllb_si(in_tmp_dir, calc):
    si = bulk('Si', 'diamond', a=5.421)
    si.calc = calc
    si.get_potential_energy()

    homo, lumo = calc.get_homo_lumo()
    response = calc.hamiltonian.xc.response
    dxc_pot = response.calculate_discontinuity_potential(homo, lumo)
    EKs, Dxc = response.calculate_discontinuity(dxc_pot)
    refgap = 3.02333
    gap = EKs + Dxc
    print('GAP', gap)
    assert gap == pytest.approx(refgap, abs=1e-4)



import pytest

import _gpaw
from gpaw.gpu import cupy as cp
from gpaw.gpu.mpi import CuPyMPI
import gpaw.mpi as mpi


@pytest.mark.gpu
def test_mpi():
    a = cp.ones(1)
    world = mpi.world
    print(a, world, a.shape, a.nbytes, a.size, a.dtype, a.data.ptr)
    if not getattr(_gpaw, 'gpu_aware_mpi', False):
        world = CuPyMPI(world)
    world.sum(a)
    assert a[0].get() == world.size



# Copyright (C) 2003  CAMP
# Copyright (C) 2010  Argonne National Laboratory
# Please see the accompanying LICENSE file for further information.

"""
Python wrapper functions for the C and Fortran packages:
Basic Linear Algebra Communication Subprogramcs (BLACS)
ScaLAPACK

See also:
http://www.netlib.org/blacs
and
http://www.netlib.org/scalapack
"""
import numpy as np

import _gpaw

switch_lu = {'U': 'L', 'L': 'U'}
switch_lr = {'L': 'R', 'R': 'L'}


def scalapack_tri2full(desc, array, conj=True):
    """Write lower triangular part into upper triangular part of matrix.

    If conj == True, the lower triangular part is the complex conjugate
    of the upper triangular part.

    This function is a frightful hack, but we can improve the
    implementation later."""

    # Zero upper triangle:
    scalapack_zero(desc, array, 'U')
    buf = array.copy()
    # Set diagonal to zero in the copy:
    scalapack_set(desc, buf, alpha=0.0, beta=0.0, uplo='U')
    # Now transpose tmp_mm adding the result to the original matrix:
    pblas_tran(alpha=1.0, a_MN=buf,
               beta=1.0, c_NM=array,
               desca=desc, descc=desc,
               conj=conj)


def scalapack_zero(desca, a, uplo, ia=1, ja=1):
    """Zero the upper or lower half of a square matrix."""
    assert desca.gshape[0] == desca.gshape[1]
    p = desca.gshape[0] - 1
    if uplo == 'L':
        ia = ia + 1
    else:
        ja = ja + 1
    scalapack_set(desca, a, 0.0, 0.0, uplo, p, p, ia, ja)


def scalapack_set(desca, a, alpha, beta, uplo, m=None, n=None, ia=1, ja=1):
    """Set the diagonal and upper/lower triangular part of a.

    Set the upper or lower triangular part of a to alpha, and the diagonal
    of a to beta, where alpha and beta are real or complex numbers."""
    desca.checkassert(a)
    assert uplo in ['L', 'U']
    if m is None:
        m = desca.gshape[0]
    if n is None:
        n = desca.gshape[1]
    if not desca.blacsgrid.is_active():
        return
    _gpaw.scalapack_set(a, desca.asarray(), alpha, beta,
                        switch_lu[uplo], n, m, ja, ia)


def scalapack_diagonalize_dc(desca, a, z, w, uplo):
    """Diagonalize symmetric matrix using the divide & conquer algorithm.
    Orthogonal eigenvectors not guaranteed; no warning is provided.

    Solve the eigenvalue equation::

      A_nn Z_nn = w_N Z_nn

    Diagonalizes A_nn and writes eigenvectors to Z_nn.  Both A_nn
    and Z_nn must be compatible with desca descriptor.  Values in
    A_nn will be overwritten.

    Eigenvalues are written to the global array w_N in ascending order.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.
    """
    desca.checkassert(a)
    desca.checkassert(z)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    assert desca.gshape[0] == len(w)
    info = _gpaw.scalapack_diagonalize_dc(a, desca.asarray(),
                                          switch_lu[uplo], z, w)
    if info != 0:
        raise RuntimeError('scalapack_diagonalize_dc error: %d' % info)


def scalapack_diagonalize_ex(desca, a, z, w, uplo, iu=None):
    """Diagonalize symmetric matrix using the bisection and inverse
    iteration algorithm. Re-orthogonalization of eigenvectors
    is an issue for tightly clustered eigenvalue problems; it
    requires substantial memory and is not scalable. See ScaLAPACK
    pdsyevx.f routine for more information.

    Solve the eigenvalue equation::

      A_nn Z_nn = w_N Z_nn

    Diagonalizes A_nn and writes eigenvectors to Z_nn.  Both A_nn
    and Z_nn must be compatible with desca descriptor.  Values in
    A_nn will be overwritten.

    Eigenvalues are written to the global array w_N in ascending order.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.

    The `iu` specifies how many eigenvectors and eigenvalues to compute.
    """
    desca.checkassert(a)
    desca.checkassert(z)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    if iu is None:  # calculate all eigenvectors and eigenvalues
        iu = desca.gshape[0]
    assert 1 < iu <= desca.gshape[0]
    # still need assert for eigenvalues
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    assert desca.gshape[0] == len(w)
    info = _gpaw.scalapack_diagonalize_ex(a, desca.asarray(),
                                          switch_lu[uplo],
                                          iu, z, w)
    if info != 0:
        # 0 means you are OK
        raise RuntimeError('scalapack_diagonalize_ex error: %d' % info)


def scalapack_diagonalize_mr3(desca, a, z, w, uplo, iu=None):
    """Diagonalize symmetric matrix using the MRRR algorithm.

    Solve the eigenvalue equation::

      A_nn Z_nn = w_N Z_nn

    Diagonalizes A_nn and writes eigenvectors to Z_nn.  Both A_nn
    and Z_nn must be compatible with this desca descriptor.  Values in
    A_nn will be overwritten.

    Eigenvalues are written to the global array w_N in ascending order.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.

    The `iu` specifies how many eigenvectors and eigenvalues to compute.
    """
    desca.checkassert(a)
    desca.checkassert(z)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    if iu is None:  # calculate all eigenvectors and eigenvalues
        iu = desca.gshape[0]
    assert 1 < iu <= desca.gshape[0]
    # stil need assert for eigenvalues
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    assert desca.gshape[0] == len(w)
    info = _gpaw.scalapack_diagonalize_mr3(a, desca.asarray(),
                                           switch_lu[uplo],
                                           iu, z, w)
    if info != 0:
        raise RuntimeError('scalapack_diagonalize_mr3 error: %d' % info)


def scalapack_general_diagonalize_dc(desca, a, b, z, w, uplo):
    """Diagonalize symmetric matrix using the divide & conquer algorithm.
    Orthogonal eigenvectors not guaranteed; no warning is provided.

    Solve the generalized eigenvalue equation::

      A_nn Z_nn = w_N B_nn Z_nn

    B_nn is assumed to be positivde definite. Eigenvectors written to Z_nn.
    Both A_nn, B_nn and Z_nn must be compatible with desca descriptor.
    Values in A_nn and B_nn will be overwritten.

    Eigenvalues are written to the global array w_N in ascending order.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.
    """
    desca.checkassert(a)
    desca.checkassert(b)
    desca.checkassert(z)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    assert desca.gshape[0] == len(w)
    info = _gpaw.scalapack_general_diagonalize_dc(a, desca.asarray(),
                                                  switch_lu[uplo], b, z, w)
    if info != 0:
        raise RuntimeError('scalapack_general_diagonalize_dc error: %d' % info)


def scalapack_general_diagonalize_ex(desca, a, b, z, w, uplo, iu=None):
    """Diagonalize symmetric matrix using the bisection and inverse
    iteration algorithm. Re-orthogonalization of eigenvectors
    is an issue for tightly clustered eigenvalue problems; it
    requires substantial memory and is not scalable. See ScaLAPACK
    pdsyevx.f routine for more information.

    Solves the eigenvalue equation::

      A_nn Z_nn = w_N B_nn Z_nn

    B_nn is assumed to be positivde definite. Eigenvectors written to Z_nn.
    Both A_nn, B_nn and Z_nn must be compatible with desca descriptor.
    Values in A_nn and B_nn will be overwritten.

    Eigenvalues are written to the global array w_N in ascending order.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.

    The `iu` specifies how many eigenvectors and eigenvalues to compute.
    """
    desca.checkassert(a)
    desca.checkassert(b)
    desca.checkassert(z)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    if iu is None:  # calculate all eigenvectors and eigenvalues
        iu = desca.gshape[0]
    assert 1 < iu <= desca.gshape[0]
    # still need assert for eigenvalues
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    assert desca.gshape[0] == len(w)
    info = _gpaw.scalapack_general_diagonalize_ex(a, desca.asarray(),
                                                  switch_lu[uplo],
                                                  iu, b, z, w)
    if info != 0:
        # 0 means you are OK
        raise RuntimeError('scalapack_general_diagonalize_ex error: %d' % info)


def scalapack_general_diagonalize_mr3(desca, a, b, z, w, uplo, iu=None):
    """Diagonalize symmetric matrix using the MRRR algorithm.

    Solve the generalized eigenvalue equation::

      A_nn Z_nn = w_N B_nn Z_nn

    B_nn is assumed to be positivde definite. Eigenvectors written to Z_nn.
    Both A_nn, B_nn and Z_nn must be compatible with desca descriptor.
    Values in A_nn and B_nn will be overwritten.

    Eigenvalues are written to the global array w_N in ascending order.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.

    The `iu` specifies how many eigenvectors and eigenvalues to compute.
    """
    desca.checkassert(a)
    desca.checkassert(b)
    desca.checkassert(z)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    if iu is None:  # calculate all eigenvectors and eigenvalues
        iu = desca.gshape[0]
    assert 1 < iu <= desca.gshape[0]
    # still need assert for eigenvalues
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    assert desca.gshape[0] == len(w)
    info = _gpaw.scalapack_general_diagonalize_mr3(a, desca.asarray(),
                                                   switch_lu[uplo],
                                                   iu, b, z, w)
    if info != 0:
        raise RuntimeError('scalapack_general_diagonalize_mr3 error: %d' %
                           info)


def scalapack_inverse_cholesky(desca, a, uplo):
    """Perform Cholesky decomposin followed by an inversion
    of the resulting triangular matrix.

    Only the upper or lower half of the matrix a will be
    modified; the other half is zeroed out.

    The `uplo` flag can be either 'L' or 'U', meaning that the
    matrices are taken to be upper or lower triangular respectively.
    """
    desca.checkassert(a)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    info = _gpaw.scalapack_inverse_cholesky(a, desca.asarray(),
                                            switch_lu[uplo])
    if info != 0:
        raise RuntimeError('scalapack_inverse_cholesky error: %d' % info)


def scalapack_inverse(desca, a, uplo):
    """Perform a hermitian matrix inversion.

    """
    desca.checkassert(a)
    # only symmetric matrices
    assert desca.gshape[0] == desca.gshape[1]
    assert uplo in ['L', 'U']
    if not desca.blacsgrid.is_active():
        return
    info = _gpaw.scalapack_inverse(a, desca.asarray(), switch_lu[uplo])
    if info != 0:
        raise RuntimeError('scalapack_inverse error: %d' % info)


def scalapack_solve(desca, descb, a, b):
    """General matrix solve.

    Solve X from A*X = B. The array b will be replaced with the result.

    This function works on the transposed form. The equivalent
    non-distributed operation is numpy.linalg.solve(a.T, b.T).T.

    This function executes the following scalapack routine:
    * pzgesv if matrices are complex
    * pdgesv if matrices are real
    """
    desca.checkassert(a)
    descb.checkassert(b)
    assert desca.gshape[0] == desca.gshape[1], 'A not a square matrix'
    assert desca.bshape[0] == desca.bshape[1], 'A not having square blocks'
    assert desca.gshape[1] == descb.gshape[1], 'B shape not compatible with A'
    assert desca.bshape[1] == descb.bshape[1], 'B blocks not compatible with A'

    if not desca.blacsgrid.is_active():
        return
    info = _gpaw.scalapack_solve(a, desca.asarray(), b, descb.asarray())
    if info != 0:
        raise RuntimeError('scalapack_solve error: %d' % info)


def pblas_tran(alpha, a_MN, beta, c_NM, desca, descc, conj=True):
    """Matrix transpose.

    C <- alpha*A.H + beta*C  if conj == True
    C <- alpha*A.T + beta*C  if conj == False

    This function executes the following PBLAS routine:
    * pztranc if matrices are complex and conj == True
    * pztranu if matrices are complex and conj == False
    * pdtran  if matrices are real
    """
    desca.checkassert(a_MN)
    descc.checkassert(c_NM)
    M, N = desca.gshape
    assert N, M == descc.gshape
    _gpaw.pblas_tran(N, M, alpha, a_MN, beta, c_NM,
                     desca.asarray(), descc.asarray(),
                     conj)


def _pblas_hemm_symm(alpha, a_MM, b_MN, beta, c_MN, desca, descb, descc,
                     side, uplo, hemm):
    """Hermitian or symmetric matrix-matrix product.

    Do not call this function directly but
    use :func:`pblas_hemm` or :func:`pblas_symm` instead.

    C <- alpha*A*B + beta*C  if side == 'L'
    C <- alpha*B*A + beta*C  if side == 'R'

    Only lower or upper diagonal of a_MM is used.

    This function executes the following PBLAS routine:
    * pzhemm if matrices are complex and hemm == True
    * pzsymm if matrices are complex and hemm == False
    * pdsymm if matrices are real
    """
    # Note: if side == 'R', then a_MM matrix is actually size of a_NN
    desca.checkassert(a_MM)
    descb.checkassert(b_MN)
    descc.checkassert(c_MN)
    assert side in ['L', 'R'] and uplo in ['L', 'U']
    Ma, Ma2 = desca.gshape
    assert Ma == Ma2, 'A not square matrix'
    Mb, Nb = descb.gshape
    if side == 'L':
        assert Mb == Ma
    else:
        assert Nb == Ma
    M, N = descc.gshape
    assert M == Mb
    assert N == Nb

    if not desca.blacsgrid.is_active():
        return
    _gpaw.pblas_hemm_symm(switch_lr[side], switch_lu[uplo],
                          N, M, alpha, a_MM, b_MN, beta, c_MN,
                          desca.asarray(), descb.asarray(), descc.asarray(),
                          hemm)


def pblas_hemm(alpha, a_MM, b_MN, beta, c_MN, desca, descb, descc,
               side='L', uplo='L'):
    """Hermitian matrix-matrix product.

    C <- alpha*A*B + beta*C  if side == 'L'
    C <- alpha*B*A + beta*C  if side == 'R'

    Only lower or upper diagonal of a_MM is used.

    This function executes the following PBLAS routine:
    * pzhemm if matrices are complex
    * pdsymm if matrices are real
    """
    return _pblas_hemm_symm(alpha, a_MM, b_MN, beta, c_MN,
                            desca, descb, descc,
                            side, uplo, hemm=True)


def pblas_symm(alpha, a_MM, b_MN, beta, c_MN, desca, descb, descc,
               side='L', uplo='L'):
    """Symmetric matrix-matrix product.

    C <- alpha*A*B + beta*C  if side == 'L'
    C <- alpha*B*A + beta*C  if side == 'R'

    Only lower or upper diagonal of a_MM is used.

    This function executes the following PBLAS routine:
    * pzsymm if matrices are complex
    * pdsymm if matrices are real
    """
    return _pblas_hemm_symm(alpha, a_MM, b_MN, beta, c_MN,
                            desca, descb, descc,
                            side, uplo, hemm=False)


def pblas_gemm(alpha, a_MK, b_KN, beta, c_MN, desca, descb, descc,
               transa='N', transb='N'):
    """General matrix-matrix product.

    C <- alpha*A*B + beta*C

    This function executes the following PBLAS routine:
    * pzgemm if matrices are complex
    * pdgemm if matrices are real
    """
    desca.checkassert(a_MK)
    descb.checkassert(b_KN)
    descc.checkassert(c_MN)
    assert transa in ['N', 'T', 'C'] and transb in ['N', 'T', 'C']
    M, Ka = desca.gshape
    Kb, N = descb.gshape

    if transa in ['T', 'C']:
        M, Ka = Ka, M
    if transb in ['T', 'C']:
        Kb, N = N, Kb
    Mc, Nc = descc.gshape

    assert Ka == Kb
    assert M == Mc
    assert N == Nc

    if not desca.blacsgrid.is_active():
        return
    _gpaw.pblas_gemm(N, M, Ka, alpha, b_KN, a_MK, beta, c_MN,
                     descb.asarray(), desca.asarray(), descc.asarray(),
                     transb, transa)


def pblas_simple_gemm(desca, descb, descc, a_MK, b_KN, c_MN,
                      transa='N', transb='N'):
    alpha = 1.0
    beta = 0.0
    pblas_gemm(alpha, a_MK, b_KN, beta, c_MN, desca, descb, descc,
               transa, transb)


def pblas_simple_hemm(desca, descb, descc, a_MM, b_MN, c_MN,
                      side='L', uplo='L'):
    alpha = 1.0
    beta = 0.0
    pblas_hemm(alpha, a_MM, b_MN, beta, c_MN, desca, descb, descc, side, uplo)


def pblas_simple_symm(desca, descb, descc, a_MM, b_MN, c_MN,
                      side='L', uplo='L'):
    alpha = 1.0
    beta = 0.0
    pblas_symm(alpha, a_MM, b_MN, beta, c_MN, desca, descb, descc, side, uplo)


def pblas_gemv(alpha, a_MN, x_N, beta, y_M, desca, descx, descy,
               transa='N'):
    """General matrix-vector product.

    y <- alpha*A*x + beta*y

    This function executes the following PBLAS routine:
    * pzgemv if matrices are complex
    * pdgemv if matrices are real
    """
    desca.checkassert(a_MN)
    descx.checkassert(x_N)
    descy.checkassert(y_M)
    assert transa in ['N', 'T', 'C']
    M, N = desca.gshape
    Nx, Ox = descx.gshape
    My, Oy = descy.gshape
    assert Ox == 1
    assert Oy == 1
    if transa == 'N':
        assert Nx == N
        assert My == M
    else:
        assert Nx == M
        assert My == N

    # Switch transposition and handle complex conjugation manually
    if transa == 'C':
        a_MN = np.ascontiguousarray(a_MN.conj())
    switch_ntc = {'N': 'T', 'T': 'N', 'C': 'N'}

    if not desca.blacsgrid.is_active():
        return
    _gpaw.pblas_gemv(N, M, alpha,
                     a_MN, x_N, beta, y_M,
                     desca.asarray(),
                     descx.asarray(),
                     descy.asarray(),
                     switch_ntc[transa])


def pblas_simple_gemv(desca, descx, descy, a, x, y, transa='N'):
    alpha = 1.0
    beta = 0.0
    pblas_gemv(alpha, a, x, beta, y, desca, descx, descy, transa)


def pblas_r2k(alpha, a_NK, b_NK, beta, c_NN, desca, descb, descc,
              uplo='U'):
    if not desca.blacsgrid.is_active():
        return
    desca.checkassert(a_NK)
    descb.checkassert(b_NK)
    descc.checkassert(c_NN)
    assert descc.gshape[0] == descc.gshape[1]  # symmetric matrix
    assert desca.gshape == descb.gshape  # same shape
    assert uplo in ['L', 'U']
    N = descc.gshape[0]  # order of C
    # K must take into account implicit tranpose due to C ordering
    K = desca.gshape[1]  # number of columns of A and B
    _gpaw.pblas_r2k(N, K, alpha, a_NK, b_NK, beta, c_NN,
                    desca.asarray(),
                    descb.asarray(),
                    descc.asarray(),
                    uplo)


def pblas_simple_r2k(desca, descb, descc, a, b, c, uplo='U'):
    alpha = 1.0
    beta = 0.0
    pblas_r2k(alpha, a, b, beta, c,
              desca, descb, descc, uplo)


def pblas_rk(alpha, a_NK, beta, c_NN, desca, descc,
             uplo='U'):
    if not desca.blacsgrid.is_active():
        return
    desca.checkassert(a_NK)
    descc.checkassert(c_NN)
    assert descc.gshape[0] == descc.gshape[1]  # symmetrix matrix
    assert uplo in ['L', 'U']
    N = descc.gshape[0]  # order of C
    # K must take into account implicit tranpose due to C ordering
    K = desca.gshape[1]  # number of columns of A
    _gpaw.pblas_rk(N, K, alpha, a_NK, beta, c_NN,
                   desca.asarray(),
                   descc.asarray(),
                   uplo)


def pblas_simple_rk(desca, descc, a, c):
    alpha = 1.0
    beta = 0.0
    pblas_rk(alpha, a, beta, c,
             desca, descc)



"""Utilities to measure and estimate memory"""

# The functions  _VmB, memory, resident, and stacksize are based on
# Python Cookbook, recipe number 286222
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/286222

import os
import numpy as np

_proc_status = '/proc/%d/status' % os.getpid()

_scale = {'kB': 1024.0, 'mB': 1024.0 * 1024.0,
          'KB': 1024.0, 'MB': 1024.0 * 1024.0}


def _VmB(VmKey):
    """Private."""
    # get pseudo file  /proc/<pid>/status
    try:
        t = open(_proc_status)
        v = t.read()
        t.close()
        # get VmKey line e.g. 'VmRSS:  9999  kB\n ...'
        i = v.index(VmKey)
    except (IOError, ValueError, LookupError):
        return 0.0  # non-Linux?

    v = v[i:].split(None, 3)  # whitespace
    if len(v) < 3:
        return 0.0  # invalid format?
    # convert Vm value to bytes
    return float(v[1]) * _scale[v[2]]


def maxrss():
    """Return maximal resident memory size in bytes."""
    # see http://www.kernel.org/doc/man-pages/online/pages/man5/proc.5.html

    # try to get it from /proc/id/status
    # This will not work on supercomputers like Blue Gene or Cray
    for name in ('VmHWM:',   # Peak resident set size ("high water mark")
                 'VmRss:',   # Resident set size
                 'VmPeak:',  # Peak virtual memory size
                 'VmSize:',  # Virtual memory size
                 ):
        mm = _VmB(name)
        if mm > 0:
            return mm

    # try to get it from rusage
    # Python documentation here:
    # http://docs.python.org/library/resource.html
    # says to multiply by the pagesize, but this is incorrect.
    # What is implementation depenedent is whether ru_maxrss is in
    # bytes or kilobytes. We make an intelligent attempt to convert
    # to detect this and convert to bytes.
    # Warning: this does not work for systems reporting kilobytes
    # for memory more than 1GiB (won't be scaled),
    # similarly for systems reporting bytes: memory of less that MiB
    # will be scaled by 1024
    #
    # the next call seems to return 'VmHWM'
    try:
        import resource
        mm = resource.getrusage(resource.RUSAGE_SELF)[2]
        if mm > 0:
            if mm < (1024)**2:  # 1 MiB
                mm = mm * 1024  # then mm was probably in KiB so convert to MiB
            return mm
    except (TypeError, ImportError):
        pass

    return 0.0  # no more ideas


class MemNode:
    """Represents the estimated memory use of an object and its components.

    Can be used on any object which implements estimate_memory().
    Example::

      from sys import stdout
      from gpaw.utilities.memory import MemNode
      node = MemNode('Root') # any name will do
      some_object.estimate_memory(node)
      nbytes = node.calculate_size()
      print('Bytes', nbytes)
      node.write(stdout) # write details

    Note that calculate_size() must be called before write().  Some
    objects must be explicitly initialized before they can estimate
    their memory use.
    """
    floatsize = np.array(1, float).itemsize
    complexsize = np.array(1, complex).itemsize
    itemsize = {float: floatsize, complex: complexsize}

    def __init__(self, name, basesize=0):
        """Create node with specified name and intrinsic size without
        subcomponents."""
        self.name = name
        self.basesize = float(basesize)
        self.totalsize = np.nan  # Size including sub-objects
        self.nodes = []
        self.indent = '    '

    def write(self, txt, maxdepth=-1, depth=0):
        """Write representation of this node and its subnodes, recursively.

        The depth parameter determines indentation.  maxdepth of -1 means
        infinity."""
        print(''.join([depth * self.indent, self.name, ': ',
                       self.memformat(self.totalsize)]), file=txt)
        if depth == maxdepth:
            return
        for node in self.nodes:
            node.write(txt, maxdepth, depth + 1)

    def memformat(self, bytes):
        # One MiB is 1024*1024 bytes, as opposed to one MB which is ambiguous
        return '%.2f MiB' % (bytes / float(1 << 20))

    def calculate_size(self):
        self.totalsize = self.basesize
        for node in self.nodes:
            self.totalsize += node.calculate_size()
        # Data-type must not be fixed-size np integer
        return self.totalsize

    def subnode(self, name, basesize=0):
        """Create subcomponent with given name and intrinsic size.  Use this
        to build component tree."""
        mem = MemNode(name, basesize)
        mem.indent = self.indent
        self.nodes.append(mem)
        return mem

    def setsize(self, basesize):
        self.basesize = float(basesize)


def monkey_patch_timer():
    """Make the timer object write information about memory usage.

    Call this function before the actual work is done.  Then do::

        $ less mem.????

    to see where the memory is allocated."""

    from ase.utils.timing import Timer

    from gpaw.mpi import rank
    from time import time

    i = Timer.__init__
    st = Timer.start
    sp = Timer.stop

    def init(self, print_levels=1000):
        i(self, print_levels)
        self.fd = open('mem.%04d' % rank, 'w')
        self.mem = []
        self.t0 = time()

    def start(self, name):
        st(self, name)
        print('start %14.6f %12d %s' % (time() - self.t0, maxrss(), name),
              file=self.fd)
        self.fd.flush()

    def stop(self, name=None):
        names = sp(self, name)
        print('stop  %14.6f %12d %s' % (time() - self.t0, maxrss(),
                                        '.'.join(names)),
              file=self.fd)
        self.fd.flush()

    Timer.__init__ = init
    Timer.start = start
    Timer.stop = stop



import os

import numpy as np

from gpaw.lcaotddft.observer import TDDFTObserver
from gpaw.tddft.units import au_to_as


class TimeDensityMatrix(TDDFTObserver):

    def __init__(self, paw, dmat, ksd, interval=1):
        TDDFTObserver.__init__(self, paw, interval)
        self.dmat = dmat
        self.ksd = ksd
        if self.ksd.world.rank == 0:
            self.dpath = 'rho'
            if not os.path.isdir(self.dpath):
                os.mkdir(self.dpath)

    def _update(self, paw):
        time = paw.time
        rho_uMM = self.dmat.get_density_matrix((paw.niter, paw.action))
        rho_up = self.ksd.transform(rho_uMM)
        if self.ksd.comm.rank == 0:
            fpath = os.path.join(self.dpath, 't%09.1f.npy' % (time * au_to_as))
            np.save(fpath, rho_up[0])



import pytest
from gpaw.mpi import serial_comm
from gpaw import GPAW
from gpaw.xc.rpa import RPACorrelation


@pytest.mark.rpa
@pytest.mark.response
def test_rpa_rpa_energy_Si(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['si_pw'], communicator=serial_comm)
    calc.diagonalize_full_hamiltonian(nbands=50)

    ecut = 50
    rpa = RPACorrelation(calc, qsym=False, nfrequencies=8, ecut=[ecut])
    E_rpa_noqsym = rpa.calculate()

    rpa = RPACorrelation(calc, qsym=True, nfrequencies=8, ecut=[ecut])
    E_rpa_qsym = rpa.calculate()

    assert E_rpa_qsym == pytest.approx(-12.61, abs=0.01)
    assert E_rpa_qsym == pytest.approx(E_rpa_noqsym)



from ase import Atoms

from gpaw import GPAW
from gpaw.mpi import serial_comm, world


def test_pw_fulldiagk(in_tmp_dir, scalapack):
    a = Atoms('H',
              cell=(1, 3, 3),
              pbc=1)

    a.calc = GPAW(mode='pw',
                  h=0.15,
                  kpts=(4, 1, 1),
                  basis='dzp',
                  nbands=4,
                  eigensolver='rmm-diis',
                  parallel={'domain': 1})

    a.get_potential_energy()
    w1 = a.calc.get_pseudo_wave_function(0, 1)
    e1 = a.calc.get_eigenvalues(1)

    a.calc.write('H.gpw')

    if world.size <= 2:
        scalapack = None
    else:
        mb = world.size // 4
        scalapack = (2, mb, 32)

    a.calc.diagonalize_full_hamiltonian(nbands=100, scalapack=scalapack)
    w2 = a.calc.get_pseudo_wave_function(0, 1)
    e2 = a.calc.get_eigenvalues(1)

    calc = GPAW('H.gpw', txt=None, parallel={'domain': 1})
    calc.diagonalize_full_hamiltonian(nbands=100, scalapack=scalapack)
    w3 = calc.get_pseudo_wave_function(0, 1)
    e3 = calc.get_eigenvalues(1)

    calc.write('Hwf.gpw', 'all')

    calc = GPAW('Hwf.gpw', txt=None, communicator=serial_comm)
    w4 = calc.get_pseudo_wave_function(0, 1)
    e4 = calc.get_eigenvalues(1)

    for w in [w2, w3, w4]:
        err = abs(abs(w[1, 2, 3]) - abs(w1[1, 2, 3]))
        print(err)
        assert err < 1e-7, err

    for e in [e2, e3, e4]:
        err = abs(e[0] - e1[0])
        assert err < 2e-9, err
        err = abs(e[-1] - e2[-1])
        assert err < 1e-10, err

    a.calc = GPAW(mode='pw',
                  h=0.15,
                  kpts=(4, 1, 1),
                  convergence={'bands': 'CBM+10'},
                  nbands=4)
    a.get_potential_energy()

    e5 = a.calc.get_eigenvalues(0)
    de = e5 - calc.get_eigenvalues(0)[:4]
    cbm = calc.get_fermi_level()
    assert abs(de[e5 < cbm + 10]).max() < 0.001



import pytest
from ase import Atoms
from gpaw import GPAW, PW
from gpaw.mpi import serial_comm
from gpaw.test import equal


@pytest.mark.response
@pytest.mark.skip(reason='TODO')
def test_rpa_C6_He():
    from gpaw.xc.rpa_correlation_energy import RPACorrelation
    ecut = 50

    He = Atoms('He')
    He.center(vacuum=1.0)

    calc = GPAW(mode=PW(force_complex_dtype=True),
                xc='PBE',
                communicator=serial_comm)
    He.calc = calc
    He.get_potential_energy()
    calc.diagonalize_full_hamiltonian()

    rpa = RPACorrelation(calc)
    C6_rpa, C6_0 = rpa.get_C6_coefficient(ecut=ecut,
                                          direction=2)

    equal(C6_0, 1.772, 0.01)
    equal(C6_rpa, 1.387, 0.01)



from __future__ import annotations

import numpy as np
from ase.units import Ha
from gpaw.core.uniform_grid import UniformGridFunctions
from gpaw.core.atom_arrays import AtomArrays
from gpaw.new import zip


class Potential:
    def __init__(self,
                 vt_sR: UniformGridFunctions,
                 dH_asii: AtomArrays,
                 energies: dict[str, float]):
        self.vt_sR = vt_sR
        self.dH_asii = dH_asii
        self.energies = energies

    def __repr__(self):
        return f'Potential({self.vt_sR}, {self.dH_asii}, {self.energies})'

    def __str__(self) -> str:
        return (f'potential:\n'
                f'  grid points: {self.vt_sR.desc.size}\n')

    def dH(self, P_ani, out_ani, spin):
        if len(P_ani.dims) == 1:  # collinear wave functions
            xp = P_ani.layout.xp
            if xp is np:
                for (a, P_ni), out_ni in zip(P_ani.items(), out_ani.values()):
                    dH_ii = self.dH_asii[a][spin]
                    np.einsum('ni, ij -> nj', P_ni, dH_ii, out=out_ni)
            else:
                for (a, P_ni), out_ni in zip(P_ani.items(), out_ani.values()):
                    dH_ii = xp.asarray(self.dH_asii[a][spin])
                    out_ni[:] = xp.einsum('ni, ij -> nj', P_ni, dH_ii)
            return  # out_ani.to_xp(to_xp)

        # Non-collinear wave functions:
        P_ansi = P_ani
        out_ansi = out_ani

        for (a, P_nsi), out_nsi in zip(P_ansi.items(), out_ansi.values()):
            v_ii, x_ii, y_ii, z_ii = (dh_ii.T for dh_ii in self.dH_asii[a])
            assert v_ii.dtype == complex
            out_nsi[:, 0] = (P_nsi[:, 0] @ (v_ii + z_ii) +
                             P_nsi[:, 1] @ (x_ii - 1j * y_ii))
            out_nsi[:, 1] = (P_nsi[:, 1] @ (v_ii - z_ii) +
                             P_nsi[:, 0] @ (x_ii + 1j * y_ii))
        return out_ansi

    def _write_gpw(self, writer, ibzwfs):
        from gpaw.new.calculation import combine_energies
        energies = combine_energies(self, ibzwfs)
        energies['band'] = ibzwfs.energies['band']
        dH_asp = self.dH_asii.to_cpu().to_lower_triangle().gather()
        vt_sR = self.vt_sR.to_xp(np).gather()
        if dH_asp is None:
            return
        writer.write(
            potential=vt_sR.data * Ha,
            atomic_hamiltonian_matrices=dH_asp.data * Ha,
            **{f'e_{name}': val * Ha for name, val in energies.items()})



import os
from pathlib import Path
from gpaw import GPAW
from ase.io import read


def test_fileio_read_old_gpw():
    dir = os.environ.get('GPW_TEST_FILES')
    if dir:
        for f in (Path(dir) / 'old').glob('*.gpw'):
            print(f)
            calc = GPAW(str(f))
            calc.get_fermi_level()
            read(f)



from .base import BaseSolver
from .bicgstab import BiCGStab
from .cscg import CSCG

__all__ = ['create_solver', 'BiCGStab', 'CSCG']


def create_solver(name, **kwargs):
    if isinstance(name, BaseSolver):
        return name
    elif isinstance(name, dict):
        kwargs.update(name)
        return create_solver(**kwargs)
    elif name == 'CSCG':
        return CSCG(**kwargs)
    elif name == 'BiCGStab':
        return BiCGStab(**kwargs)
    else:
        raise ValueError('Unknown solver: %s' % name)
